"""This class groups the methods related to the Inter AAS interactions between I4.0 SMIA entities."""
import json
import logging

from smia.logic import acl_smia_messages_utils
from smia.utilities.fipa_acl_info import FIPAACLInfo
from spade.message import Message



_logger = logging.getLogger(__name__)


def create_svc_json_data_from_acl_msg(acl_msg):
    """
    This method creates the dictionary with all the required data of a service related to an ACL message.

    Args:
        acl_msg (spade.message.Message): ACL message where to get the information

    Returns:
        dict: dictionary with all the information about the service
    """
    svc_req_data_json = {
        'performative': acl_msg.get_metadata('performative'),
        'ontology': acl_msg.get_metadata('ontology'),
        'thread': acl_msg.thread,
        'sender': acl_smia_messages_utils.get_sender_from_acl_msg(acl_msg),
        'receiver': str(acl_msg.to),
    }
    # The body of the ACL message contains the rest of the information
    svc_req_data_json.update(json.loads(acl_msg.body))
    return svc_req_data_json

# Methods related to the ACL-SMIA messages
# ----------------------------------------
def create_inter_smia_response_msg(receiver, thread, performative, ontology, service_id=None, service_type=None,
                                   service_params=None):
    """
    This method creates the Inter AAS interaction response object.

    Args:
        receiver (str): the JID of the receiver of the ACL message from which the service is requested.
        thread (str): the thread of the ACL message.
        performative (str): the performative of the ACL message.
        ontology (str): the ontology of the ACL message.
        service_id (str): the serviceID of the ACL message.
        service_type (str): the serviceType of the ACL message.
        service_params (str): the serviceParams of the "serviceData" section of the ACL message.

    Returns:
        spade.message.Message: SPADE message object FIPA-ACL-compliant.
    """
    from smia.utilities.general_utils import GeneralUtils  # Local imports to avoid circular import error

    request_msg = Message(to=receiver, thread=thread)
    request_msg.set_metadata('performative', performative)
    request_msg.set_metadata('ontology', ontology)
    # request_msg.set_metadata('ontology', 'SvcResponse')

    request_msg_body_json = {
        'serviceID': service_id,
        'serviceType': service_type,
        'serviceData': {
            'serviceCategory': 'service-response',
            'timestamp': GeneralUtils.get_current_timestamp(),
        }
    }
    if service_params is not None:
        request_msg_body_json['serviceData']['serviceParams'] = service_params
    request_msg.body = json.dumps(request_msg_body_json)
    return request_msg

async def create_acl_response_from_received_msg(received_msg, performative, response_body=None):
    """
    This method creates an Inter SMIA interaction response object from a received ACL message. Thus, some of the
    required data will be obtained from the received message (receiver, thread, ontology, protocol, encoding and
    language).

    Args:
        received_msg (spade.message.Message): the received ACL message.
        performative (str): the performative of the ACL message.
        response_body: the body of the ACL response message.

    Returns:
        spade.message.Message: SPADE message object FIPA-ACL-compliant.
    """

    # The response message is built with the data from the received message, adding the desired performative
    response_msg = Message(to=acl_smia_messages_utils.get_sender_from_acl_msg(received_msg), thread=received_msg.thread)
    response_msg.set_metadata(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_ATTRIB, performative)
    response_msg.set_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB,
                              received_msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB))
    if received_msg.get_metadata(FIPAACLInfo.FIPA_ACL_PROTOCOL_ATTRIB) is not None:
        response_msg.set_metadata(FIPAACLInfo.FIPA_ACL_PROTOCOL_ATTRIB,
                                  received_msg.get_metadata(FIPAACLInfo.FIPA_ACL_PROTOCOL_ATTRIB))
    if received_msg.get_metadata(FIPAACLInfo.FIPA_ACL_ENCODING_ATTRIB) is not None:
        response_msg.set_metadata(FIPAACLInfo.FIPA_ACL_ENCODING_ATTRIB,
                                  received_msg.get_metadata(FIPAACLInfo.FIPA_ACL_ENCODING_ATTRIB))
    if received_msg.get_metadata(FIPAACLInfo.FIPA_ACL_LANGUAGE_ATTRIB) is not None:
        response_msg.set_metadata(FIPAACLInfo.FIPA_ACL_LANGUAGE_ATTRIB,
                                  received_msg.get_metadata(FIPAACLInfo.FIPA_ACL_LANGUAGE_ATTRIB))

    if response_body is not None:
        if isinstance(response_body, dict):
            response_msg.body = json.dumps(response_body)
        else:
            response_msg.body = str(response_body)
    return response_msg


async def create_acl_smia_message(receiver, thread, performative, ontology, msg_body=None, protocol=None, encoding=None,
                            language=None):
    """
    This method creates a FIPA-ACL-SMIA message for an Inter SMIA interaction. If optional attributes are set, they will
     be added to the message.

    Args:
        receiver (str): the JID of the receiver of the ACL message from which the service is requested.
        thread (str): the thread of the ACL message.
        performative (str): the performative of the ACL message.
        ontology (str): the ontology of the ACL message.
        msg_body: the vody of the ACL message.
        protocol (str): the protocol of the ACL message.
        encoding (str): the encoding of the ACL message.
        language (str): the language of the ACL message.

    Returns:
        spade.message.Message: SPADE message object FIPA-ACL-SMIA-compliant.
    """

    # The response message is built with the data from the received message, adding the desired performative
    acl_smia_msg = Message(to=receiver, thread=thread)
    acl_smia_msg.set_metadata(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_ATTRIB, performative)
    acl_smia_msg.set_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB, ontology)
    # Optional attributes are added if they are set
    if protocol is not None:
        acl_smia_msg.set_metadata(FIPAACLInfo.FIPA_ACL_PROTOCOL_ATTRIB, protocol)
    if encoding is not None:
        acl_smia_msg.set_metadata(FIPAACLInfo.FIPA_ACL_ENCODING_ATTRIB, encoding)
    elif encoding is None:
        acl_smia_msg.set_metadata(FIPAACLInfo.FIPA_ACL_ENCODING_ATTRIB, FIPAACLInfo.FIPA_ACL_DEFAULT_ENCODING)
    if language is not None:
        acl_smia_msg.set_metadata(FIPAACLInfo.FIPA_ACL_LANGUAGE_ATTRIB, language)
    elif encoding is None:
        acl_smia_msg.set_metadata(FIPAACLInfo.FIPA_ACL_LANGUAGE_ATTRIB, FIPAACLInfo.FIPA_ACL_DEFAULT_LANGUAGE)

    if msg_body is not None:
        if isinstance(msg_body, dict):
            acl_smia_msg.body = json.dumps(msg_body)
        else:
            acl_smia_msg.body = str(msg_body)
    return acl_smia_msg

async def acl_message_to_json(acl_message):
    """
    This method converts a FIPA-ACL-SMIA message to JSON object.
    Args:
        acl_message (spade.message.Message): SPADE message object FIPA-ACL-SMIA-compliant.

    Returns:
        dict: JSON object with all the information of the ACL message.
    """
    return {
        'sender': acl_smia_messages_utils.get_sender_from_acl_msg(acl_message),
        'receiver': str(acl_message.to),
        'thread': acl_message.thread,
        FIPAACLInfo.FIPA_ACL_PERFORMATIVE_ATTRIB: acl_message.get_metadata(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_ATTRIB),
        FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB: acl_message.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB),
        FIPAACLInfo.FIPA_ACL_PROTOCOL_ATTRIB: acl_message.get_metadata(FIPAACLInfo.FIPA_ACL_PROTOCOL_ATTRIB),
        FIPAACLInfo.FIPA_ACL_ENCODING_ATTRIB: acl_message.get_metadata(FIPAACLInfo.FIPA_ACL_ENCODING_ATTRIB),
        FIPAACLInfo.FIPA_ACL_LANGUAGE_ATTRIB: acl_message.get_metadata(FIPAACLInfo.FIPA_ACL_LANGUAGE_ATTRIB),
        'body': json.loads(acl_message.body),
    }

# Methods related to the body of ACL-SMIA messages
# ------------------------------------------------
async def check_received_request_data_structure_old(received_data, json_schema):
    """
    This method checks if the received data for a request is valid. The JSON object with the specific
    data is also validated against the given associated JSON Schema. In any case, if it is invalid, it raises a
    RequestDataError exception.

    Args:
        received_data (dict): received data in form of a JSON object.
        json_schema (dict): JSON Schema in form of a JSON object.
    """
    import jsonschema
    from smia.logic.exceptions import RequestDataError
    from jsonschema.exceptions import ValidationError
    # TODO modificarlo cuando se piense la estructura del lenguaje I4.0
    if 'serviceData' not in received_data:
        raise RequestDataError("The received request is invalid due to missing #serviceData field in the"
                               "request message.")
    if 'serviceParams' not in received_data['serviceData']:
        raise RequestDataError("The received request is invalid due to missing #serviceParams field within "
                               "the #serviceData section of the request message.")
    # The received JSON object is also validated against the associated JSON Schema
    try:
        jsonschema.validate(instance=received_data['serviceData']['serviceParams'],
                            schema=json_schema)
    except ValidationError as e:
        raise RequestDataError("The received JSON data within the request message is invalid against the required "
                               "JSON schema. Invalid part: {}. Reason: {}.".format(e.instance, e.message))

async def check_received_request_data_structure(received_data, json_schema):
    """
    This method checks if the received data for a request is valid. So, the JSON object with the specific
    data is validated against the given associated JSON Schema for FIPA-ACL-SMIA messages. In any case, if it is
    invalid, it raises a RequestDataError exception.

    Args:
        received_data (dict): received data in form of a JSON object.
        json_schema (dict): JSON Schema in form of a JSON object.
    """
    import jsonschema
    from jsonschema.exceptions import ValidationError
    from smia.logic.exceptions import RequestDataError  # Local imports to avoid circular import error

    # The received JSON object is validated against the associated JSON Schema
    try:
        jsonschema.validate(instance=received_data, schema=json_schema)
    except ValidationError as e:
        raise RequestDataError("The received JSON data within the request message is invalid against the required "
                               "JSON schema. Invalid part: {}. Reason: {}.".format(e.instance, e.message))

# Methods related to SMIA instances interactions
# ----------------------------------------------
async def send_response_msg_from_received(agent_behav, received_msg, performative, response_body=None):
    """
    This method sends a response message from a received one, adding the desired data: performative and body.

    Args:
        agent_behav (AgentBehavior): Agent behaviour object.
        received_msg (spade.message.Message): the received ACL message.
        performative (str): the performative of the ACL message.
        response_body: the body of the ACL response message.
    """
    response_msg = await create_acl_response_from_received_msg(received_msg, performative, response_body)
    await agent_behav.send(response_msg)



import json
import logging

from basyx.aas.adapter.json import AASToJsonEncoder
from spade.behaviour import OneShotBehaviour

from smia import GeneralUtils
from smia.aas_model.aas_model_utils import AASModelUtils
from smia.logic import inter_smia_interactions_utils, acl_smia_messages_utils
from smia.logic.exceptions import RequestDataError, ServiceRequestExecutionError, AASModelReadingError, \
    AssetConnectionError
from smia.logic.services_utils import AgentServiceUtils
from smia.utilities import smia_archive_utils
from smia.utilities.aas_related_services_info import AASRelatedServicesInfo
from smia.utilities.fipa_acl_info import FIPAACLInfo, ServiceTypes, ACLSMIAOntologyInfo
from smia.utilities.smia_info import AssetInterfacesInfo

_logger = logging.getLogger(__name__)


class HandleAASRelatedSvcBehaviour(OneShotBehaviour):
    """
    This class implements the behaviour that handles all the service requests that the SMIA has received. This
    request can arrive from an FIPA-ACL message as a :term:`Inter AAS Interaction` or from the AAS Core as an
    :term:`Intra AAS Interaction` message. This is a OneShotBehaviour because it handles an individual service request
    and then kills itself.
    """

    # TODO PENSAR SI AGRUPAR EN ESTA CLASE TANTO Requests como Responses ('HandleSvcBehaviour'), ya que para CSS solo
    #  hay CapabilityBehaviour. Dentro de esta se podria analizar la performativa (como ya se hace), para ver si es una
    #  peticion o una respuesta

    def __init__(self, agent_object, received_acl_msg):
        """
        The constructor method is rewritten to add the object of the agent.

        Args:
            agent_object (spade.Agent): the SPADE agent object of the SMIA agent.
            received_acl_msg (spade.message.Message): the received ACL-SMIA message object
        """

        # The constructor of the inherited class is executed.
        super().__init__()

        # The SPADE agent object is stored as a variable of the behaviour class
        self.myagent = agent_object
        self.received_acl_msg = received_acl_msg
        self.received_body_json = acl_smia_messages_utils.get_parsed_body_from_acl_msg(self.received_acl_msg)

        self.requested_timestamp = GeneralUtils.get_current_timestamp()

    async def on_start(self):
        """
        This method implements the initialization process of this behaviour.
        """
        _logger.info("HandleAASRelatedSvcBehaviour starting to handle the service related to the message with thread"
                     " [{}]...".format(self.received_acl_msg.thread))

    async def run(self):
        """
        This method implements the logic of the behaviour.
        """
        # Depending on the type of service, the associated method will be launched
        match self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB):
            case (ACLSMIAOntologyInfo.ACL_ONTOLOGY_ASSET_RELATED_SERVICE |
                  ACLSMIAOntologyInfo.ACL_ONTOLOGY_AGENT_RELATED_SERVICE):
                await self.handle_asset_agent_related_service()
            case ACLSMIAOntologyInfo.ACL_ONTOLOGY_AAS_SERVICE:
                await self.handle_aas_service()
            case ACLSMIAOntologyInfo.ACL_ONTOLOGY_AAS_INFRASTRUCTURE_SERVICE:
                await self.handle_aas_infrastructure_service()

        _logger.info("Management of the AAS-related service with thread {} finished.".format(self.received_acl_msg.thread))

        # TODO BORRAR ENFOQUE ANTIGUO
        # # First, the performative of the request is obtained
        # match self.svc_req_data['performative']:
        #     case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_REQUEST:  # TODO actualizar dentro de todo el codigo los usos de performativas y ontologias de FIPA-ACL
        #         await self.handle_request()
        #     case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_QUERY_IF:
        #         await self.handle_query_if()
        #     # TODO PensarOtros
        #     case _:
        #         _logger.error("Performative not available for service management.")

    # -------------------------------------
    # Asset-/agent-related services methods
    # -------------------------------------
    async def handle_asset_agent_related_service(self):
        """
        This method implements the logic to handle the asset- or agent-related type services.
        """
        try:
            match self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_ATTRIB):
                case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_REQUEST:
                    result = await self.handle_asset_agent_related_svc_request()
                case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM:
                    _logger.aclinfo(f"The SMIA has been informed about the asset-/agent-related service related to the "
                                    f"thread [{self.received_acl_msg.thread}] with the content:{self.received_acl_msg.body}.")
                    # TODO
                    return
                case _:
                    unsupported_performative_msg = ("Cannot handle the asset-/agent-related service of the ACL "
                                                    "interaction with thread [{}] because the performative [{}] is not"
                                                    " yet supported.".format(
                        self.received_acl_msg.thread,
                        self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_ATTRIB)))
                    _logger.warning(unsupported_performative_msg)
                    raise RequestDataError(unsupported_performative_msg)

            # When the service is successfully performed the result will be sent to the requester
            await inter_smia_interactions_utils.send_response_msg_from_received(
                    self, self.received_acl_msg, FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM, response_body=result)

            # The information will be stored in the log
            # TODO MODIFICAR ESTO CON LAS NUEVAS ESTRUCTURAS DE MENSAJES JSON
            smia_archive_utils.save_completed_svc_log_info(
                self.requested_timestamp, GeneralUtils.get_current_timestamp(),
                await inter_smia_interactions_utils.acl_message_to_json(self.received_acl_msg), str(result),
                self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB))
        except (RequestDataError, ServiceRequestExecutionError,
                AASModelReadingError, AssetConnectionError) as svc_request_error:

            if isinstance(svc_request_error, RequestDataError):
                svc_request_error = ServiceRequestExecutionError(
                    self.received_acl_msg.thread, svc_request_error.message,
                    self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB), self)
            if isinstance(svc_request_error, AASModelReadingError):
                svc_request_error = ServiceRequestExecutionError(
                    self.received_acl_msg.thread, "{}. Reason:  {}".format(svc_request_error.message,
                                                                           svc_request_error.reason),
                    self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB), self)
            if isinstance(svc_request_error, AssetConnectionError):
                svc_request_error = ServiceRequestExecutionError(
                    self.received_acl_msg.thread,f"The error [{svc_request_error.error_type}] has appeared "
                                                 f"during the asset connection. Reason: {svc_request_error.reason}.",
                    self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB), self)

            await svc_request_error.handle_service_execution_error()
            return  # killing a behaviour does not cancel its current run loop

    async def handle_asset_agent_related_svc_request(self):
        """
        This method handles an Asset Related Service request. These services are part of I4.0 Application Component
        (application relevant).
        """
        # The asset/agent related service will be executed using a ModelReference to the related SubmodelElement related
        # with the execution method (in case of agent service) or element within AssetInterfacesSubmodel (in case of
        # asset service).

        # Since at this point the received data is valid, the AAS Reference object need to be created
        if isinstance(self.received_body_json['serviceRef'], str):
            self.received_body_json['serviceRef'] = await AASModelUtils.aas_model_reference_string_to_dict(
                self.received_body_json['serviceRef'])
        aas_asset_agent_service_ref = await AASModelUtils.create_aas_reference_object(
            'ModelReference', self.received_body_json['serviceRef'])
        aas_asset_agent_service_elem = await self.myagent.aas_model.get_object_by_reference(aas_asset_agent_service_ref)

        # If serviceParams is not defined in the JSON body is added None value to avoid errors
        if 'serviceParams' not in self.received_body_json:
            self.received_body_json['serviceParams'] = None

        # The asset connection class can be used to know the type of the service and it is also required to execute the
        # asset related service. It is obtained from the AAS asset service Reference object
        aas_asset_interface_ref = aas_asset_agent_service_elem.get_parent_ref_by_semantic_id(
            AssetInterfacesInfo.SEMANTICID_INTERFACE)
        if aas_asset_interface_ref is None:
            # In this case it is an agent-related service
            adapted_svc_params = await AgentServiceUtils.adapt_received_service_parameters(
                await self.myagent.agent_services.get_agent_service_by_id(aas_asset_agent_service_elem.id_short),
                self.received_body_json['serviceParams'])
            _logger.info(f"Executing agent service [{aas_asset_agent_service_elem.id_short}] "
                              f"within agent-related service of thread [{self.received_acl_msg.thread}]...")
            agent_service_execution_result = await self.myagent.agent_services.execute_agent_service_by_id(
                aas_asset_agent_service_elem.id_short, **adapted_svc_params)
            _logger.info(f"Successfully executed agent service [{aas_asset_agent_service_elem.id_short}] "
                              f"within agent-related service of thread [{self.received_acl_msg.thread}]...")
            return agent_service_execution_result
        else:
            # TODO PROBARLO (aunque deberia funcionar)
            asset_connection_class = await self.myagent.get_asset_connection_class_by_ref(
                aas_asset_interface_ref)

            # With all necessary information obtained, the asset related service can be executed
            _logger.assetinfo(f"Executing asset service [{aas_asset_agent_service_elem.id_short}] "
                              f"within asset-related service of thread [{self.received_acl_msg.thread}]...")
            asset_service_execution_result = await asset_connection_class.execute_asset_service(
                interaction_metadata=aas_asset_agent_service_elem,
                service_input_data=self.received_body_json['serviceParams'])
            _logger.assetinfo(f"Executing asset service [{aas_asset_agent_service_elem.id_short}] "
                              f"within asset-related service of thread [{self.received_acl_msg.thread}]...")
            return asset_service_execution_result

    # --------------------
    # AAS services methods
    # --------------------
    async def handle_aas_service(self):
        """
        This method implements the logic to handle the AAS type services.
        """
        try:
            match self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_ATTRIB):
                case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_QUERY_REF:
                    result = await self.handle_aas_svc_query_ref()
                case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM:
                    _logger.aclinfo(f"The SMIA has been informed about the AAS service related to the thread"
                                    f" [{self.received_acl_msg.thread}] with the content:{self.received_acl_msg.body}.")
                    # TODO
                    return
                case _:
                    unsupported_performative_msg = ("Cannot handle the asset-/agent-related service of the ACL "
                                                    "interaction with thread [{}] because the performative [{}] is not"
                                                    " yet supported.".format(
                        self.received_acl_msg.thread,
                        self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_ATTRIB)))
                    _logger.warning(unsupported_performative_msg)
                    raise RequestDataError(unsupported_performative_msg)

            # When the service is successfully performed the result will be sent to the requester
            await inter_smia_interactions_utils.send_response_msg_from_received(
                self, self.received_acl_msg, FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM, response_body=result)

            # The information will be stored in the log
            # TODO MODIFICAR ESTO CON LAS NUEVAS ESTRUCTURAS DE MENSAJES JSON
            smia_archive_utils.save_completed_svc_log_info(
                self.requested_timestamp, GeneralUtils.get_current_timestamp(),
                await inter_smia_interactions_utils.acl_message_to_json(self.received_acl_msg), str(result),
                self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB))
        except (RequestDataError, ServiceRequestExecutionError, AASModelReadingError) as svc_request_error:
            if isinstance(svc_request_error, RequestDataError):
                svc_request_error = ServiceRequestExecutionError(
                    self.received_acl_msg.thread, svc_request_error.message,
                    self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB), self)
            if isinstance(svc_request_error, AASModelReadingError):
                svc_request_error = ServiceRequestExecutionError(
                    self.received_acl_msg.thread, "{}. Reason: {}".format(svc_request_error.message,
                                                                          svc_request_error.reason),
                    self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB), self)
            await svc_request_error.handle_service_execution_error()
            return  # killing a behaviour does not cancel its current run loop

    async def handle_aas_svc_query_ref(self):
        """
        This method implements the logic to handle the Query-Ref performatives of AAS type services.
        """
        match self.received_body_json['serviceType']:
            case AASRelatedServicesInfo.AAS_SERVICE_TYPE_DISCOVERY:
                return await self.handle_aas_discovery_svc()
            case _:
                unsupported_service_type = ("Cannot handle the AAS service of the ACL interaction with thread [{}] "
                                            "because the service type [{}] is not yet supported.".format(
                    self.received_acl_msg.thread, self.received_body_json['serviceType']))
                _logger.warning(unsupported_service_type)
                raise RequestDataError(unsupported_service_type)

    async def handle_aas_discovery_svc(self):
        """
        This method implements the logic to handle the AAS type services of Discovery ServiceTypes.
        """
        match self.received_body_json['serviceID']:
            case (AASRelatedServicesInfo.AAS_DISCOVERY_SERVICE_GET_SM_BY_REF |
                  AASRelatedServicesInfo.AAS_DISCOVERY_SERVICE_GET_SM_VALUE_BY_REF):

                if isinstance(self.received_body_json['serviceParams'], str):
                    self.received_body_json['serviceParams'] = await AASModelUtils.aas_model_reference_string_to_dict(
                        self.received_body_json['serviceParams'])

                self.received_body_json['serviceParams'] = await self.myagent.aas_model.check_and_adapt_for_templates(
                    self.received_body_json['serviceParams'])

                aas_object_ref = await AASModelUtils.create_aas_reference_object(
                    'ModelReference', self.received_body_json['serviceParams'])



                # When the appropriate Reference object is created, the requested SubmodelElement can be obtained
                requested_sme = await self.myagent.aas_model.get_object_by_reference(aas_object_ref)

                # When the AAS object has been obtained, the result is returned in JSON format
                if self.received_body_json['serviceID'] == AASRelatedServicesInfo.AAS_DISCOVERY_SERVICE_GET_SM_BY_REF:
                    return json.dumps(requested_sme, cls=AASToJsonEncoder)
                else:
                    if not hasattr(requested_sme, 'value'):
                        raise RequestDataError(f"The SubmodelElement queried with the reference ["
                                               f"{self.received_body_json['serviceParams']}] has not value, so it cannot"
                                               f" be returned")
                    if not isinstance(requested_sme.value, str):
                        return json.dumps(requested_sme.value, cls=AASToJsonEncoder)
                    return str(requested_sme.value)
            case AASRelatedServicesInfo.AAS_DISCOVERY_SERVICE_GET_AAS_INFO:
                #TODO
                pass
            case AASRelatedServicesInfo.AAS_DISCOVERY_SERVICE_GET_SM_BY_ID:
                # TODO
                pass
            case _:
                unsupported_service_id = ("Cannot handle the AAS service of the ACL interaction with thread [{}] "
                                            "because the discovery service id [{}] is not yet supported.".format(
                    self.received_acl_msg.thread, self.received_body_json['serviceID']))
                _logger.warning(unsupported_service_id)
                raise RequestDataError(unsupported_service_id)

    # -----------------------------------
    # AAS infrastructure services methods
    # -----------------------------------
    async def handle_aas_infrastructure_service(self):
        """
        This method implements the logic to handle the Infrastructure services of AAS type services.
        """
        # The AAS Infrastructure Services are offered by the platform (in SMIA approach by SMIA ISM), not by SMIA
        # instances, so it will not realize any task
        pass

    # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # TODO BORRAR ENFOQUE ANTIGUO
    # ------------------------------------------
    # Methods to handle of all types of services
    # ------------------------------------------
    async def handle_request(self):
        """
        This method handle capability requests to the DT.
        """
        # The type is analyzed to perform the appropriate service
        match self.svc_req_data['serviceType']:
            case ServiceTypes.ASSET_RELATED_SERVICE:
                # await self.handle_asset_related_svc()   # TODO
                await self.handle_asset_agent_related_svc_request()
            case ServiceTypes.AAS_INFRASTRUCTURE_SERVICE:
                await self.handle_aas_infrastructure_svc_request()  # TODO
            case ServiceTypes.AAS_SERVICE:
                await self.handle_aas_services_request()  # TODO
            case ServiceTypes.SUBMODEL_SERVICE:
                await self.handle_submodel_service_request()
            case _:
                _logger.error("Service type not available.")

    async def handle_query_if(self):
        """This method handle Query-If service requests. This request is received when the DT is asked about information
         related to a service."""
        pass
        # TODO FALTA POR HACER

    # -----------------------------------
    # Methods to handle specific services
    # -----------------------------------
    async def handle_aas_services_request(self):
        """
        This method handles AAS Services. These services serve for the management of asset-related information through
        a set of infrastructure services provided by the AAS itself. These include Submodel Registry Services (to list
        and register submodels), Meta-information Management Services (including Classification Services, to check if the
        interface complies with the specifications; Contextualization Services, to check if they belong together in a
        context to build a common function; and Restriction of Use Services, divided between access control and usage
        control) and Exposure and Discovery Services (to search for submodels or asset related services).

        """
        _logger.info('AAS Service request: ' + str(self.svc_req_data))

    async def handle_aas_infrastructure_svc_request(self):
        """
        This method handles AAS Infrastructure Services. These services are part of I4.0 Infrastructure Services
        (Systemic relevant). They are necessary to create AASs and make them localizable and are not offered by an AAS, but
        by the platform (computational infrastructure). These include the AAS Create Service (for creating AASs with unique
        identifiers), AAS Registry Services (for registering AASs) and AAS Exposure and Discovery Services (for searching
        for AASs).

        """
        _logger.info('AAS Infrastructure Service request: ' + str(self.svc_req_data))




    async def send_response_msg_to_sender(self, performative, service_params):
        """
        This method creates and sends a FIPA-ACL message with the given serviceParams and performative.

        Args:
            performative (str): performative according to FIPA-ACL standard.
            service_params (dict): JSON with the serviceParams to be sent in the message.
        """
        acl_msg = inter_smia_interactions_utils.create_inter_smia_response_msg(
            receiver=self.svc_req_data['sender'],
            thread=self.svc_req_data['thread'],
            performative=performative,
            ontology='SvcResponse',
            service_id=self.svc_req_data['serviceID'],
            service_type=self.svc_req_data['serviceType'],
            service_params=json.dumps(service_params)
        )
        await self.send(acl_msg)




import asyncio
import logging
import random

from spade.behaviour import CyclicBehaviour

from smia import GeneralUtils
from smia.css_ontology import css_operations
from smia.logic import negotiation_utils, inter_smia_interactions_utils, acl_smia_messages_utils
from smia.logic.exceptions import CapabilityRequestExecutionError, AssetConnectionError
from smia.utilities import smia_archive_utils
from smia.utilities.fipa_acl_info import FIPAACLInfo, ACLSMIAOntologyInfo
from smia.utilities.smia_info import AssetInterfacesInfo

_logger = logging.getLogger(__name__)


class HandleNegotiationBehaviour(CyclicBehaviour):
    """
    This class implements the behaviour that handle a particular negotiation.
    """
    myagent = None  #: the SPADE agent object of the SMIA agent.
    neg_value = None  #: value of the negotiation
    targets_processed = set()  #: targets that their values have been processed
    neg_value_event = None

    def __init__(self, agent_object, received_acl_msg):
        """
        The constructor method is rewritten to add the object of the agent
        Args:
            agent_object (spade.Agent): the SPADE agent object of the SMIA agent.
            received_acl_msg (spade.message.Message): the received ACL-SMIA message object
        """

        # The constructor of the inherited class is executed.
        super().__init__()

        # The SPADE agent object is stored as a variable of the behaviour class
        self.myagent = agent_object

        self.received_acl_msg = received_acl_msg
        self.received_body_json = acl_smia_messages_utils.get_parsed_body_from_acl_msg(self.received_acl_msg)

        if 'negRequester' not in self.received_body_json:
            # If it has not been added, it is added using the sender identifier
            self.received_body_json['negRequester'] = acl_smia_messages_utils.get_sender_from_acl_msg(
                self.received_acl_msg)

        # Negotiation-related variables are also initialized
        self.neg_thread = self.received_acl_msg.thread
        self.all_targets_list = set(self.received_body_json['negTargets'])
        if str(self.myagent.jid) in self.all_targets_list:
            self.all_targets_list.remove(str(self.myagent.jid))   # Removes its own JID
        self.targets_processed = set()
        self.targets_remaining = set(self.received_body_json['negTargets'])

        self.neg_value = 0.0
        self.negotiation_result = None

        # If not all proposals have been received, three retries are set to request them again (using request)
        self.max_retries = 3
        self.current_retries = 0

        # The safe iterations (to keep the behavior alive in case the other participants have not finished processing)
        # are set depending on the number of participants in the negotiation. The total number of iterations is at
        # least 10 or based on the number of targets
        self.safe_iterations = max(10, int(len(self.received_body_json['negTargets'])*0.2))

        self.requested_timestamp = GeneralUtils.get_current_timestamp()

    async def on_start(self):
        """
        This method implements the initialization process of this behaviour.
        """
        _logger.info("HandleNegotiationBehaviour starting for [{}]...".format(self.neg_thread))

        # Since this behavior is specific to the messages in this thread, it reserves it so that the generic
        # ACLHandlingBehavior does not process them.
        await self.myagent.add_reserved_thread(self.neg_thread)

        # Before starting, if the FIPA-CNP protocol is related to a CSS service, it will perform the capability checking
        if (self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB) ==
                ACLSMIAOntologyInfo.ACL_ONTOLOGY_CSS_SERVICE):
            result, reason = await css_operations.capability_checking(self.myagent, self.received_body_json)
            if not result:
                _logger.info("The SMIA has received a negotiation with failed capability checking [" +
                             self.neg_thread + "]")

                # Since the capability check failed, it will reply to the sender with a Refuse message
                await inter_smia_interactions_utils.send_response_msg_from_received(
                    self, self.received_acl_msg, FIPAACLInfo.FIPA_ACL_PERFORMATIVE_REFUSE,
                    response_body={'reason': 'CapabilityChecking failed: {}'.format(reason)})
                _logger.aclinfo("ACL response sent for the result of the negotiation request with thread ["
                                + self.neg_thread + "]")

                # The negotiation value will be -1 to lose all comparisons
                self.neg_value = -1.0

        # First, it will analyze whether it is the only participant in the negotiation, in which case it will be the
        # direct winner
        if len(self.received_body_json['negTargets']) == 1:
            # There is only one target available (therefore, it is the only one, so it is the winner)
            _logger.info("The SMIA has won the negotiation with thread [" + self.neg_thread + "]")

            # As the winner, it will reply to the sender with the result of the negotiation
            await inter_smia_interactions_utils.send_response_msg_from_received(
                self, self.received_acl_msg, FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM, response_body={'winner': True})
            _logger.aclinfo("ACL response sent for the result of the negotiation request with thread ["
                            + self.neg_thread + "]")

            # The information will be stored in the log and the SMIA instance ends as the winner
            await self.exit_negotiation(is_winner=True)

        else:
            # In this case, there are multiple participants, so it will execute the FIPA-SMIA-CNP protocol
            try:
                # First, the times between message transmissions and active waiting times for reception are calculated.
                # Dynamic wait time between messages: wait time to allow incoming messages to be processed while other
                # messages are being sent.
                time_base = 0.005   # is set as the minimum time to serialize an XMPP message.
                time_jitter = 0.001 # Randomness to avoid scenarios where everyone sends at the same time
                time_sleep_base = time_base + (len(self.received_body_json['negTargets']) * time_jitter)
                time_sleep_range = time_sleep_base * 0.1    # 10 %
                self.time_sleep = time_sleep_base + random.uniform(-time_sleep_range, 0)

                # Dynamic recovery timeout: waiting time to process messages to be received once all messages have been
                # sent (e.g., when sending retry requests)
                security_factor = 3.0   # A multiplier to absorb network delays
                total_transmission_time = (len(self.received_body_json['negTargets']) - 1) * self.time_sleep
                self.recovery_wait = total_transmission_time * security_factor

                #  The value of the criterion must be obtained just before starting to manage the negotiation, so that at the
                #  time of sending the PROPOSE and receiving that of the others it will be the same value.
                await self.get_neg_value_with_criteria()

                # Once the negotiation value is reached, the negotiation management can begin (run method).
            except (CapabilityRequestExecutionError, AssetConnectionError) as cap_neg_error:
                if isinstance(cap_neg_error, AssetConnectionError):
                    cap_neg_error = CapabilityRequestExecutionError(self.neg_thread,'Negotiation',
                                                                    f"The error [{cap_neg_error.error_type}] "
                                                                    f"has appeared during the asset connection. "
                                                                    f"Reason: {cap_neg_error.reason}.", self)

                await cap_neg_error.handle_capability_execution_error_old()
                return  # killing a behaviour does not cancel its current run loop


    async def run(self):
        """
        This method implements the logic of the behaviour.
        """
        # if self.initial_iteration:
        #     # In the first iteration, each agent waits for a set amount of time relative to the number of participants
        #     # in the negotiation to ensure that everyone has arrived at their run method.
        #     await asyncio.sleep(max(3, 0.05*len(self.all_targets_list)))
        #     self.initial_iteration = False

        # The first step in each iteration is to send the PROPOSE message with your own value to the other participants
        # in the negotiation. In this case, one iteration is performed to distribute the communication load.
        await self.send_propose_acl_msgs()  # Envia los propose en la primera iteracion de espera

        if len(self.targets_remaining) == 0:
            timeout = self.recovery_wait
            # await asyncio.sleep(self.recovery_wait)
        else:
            timeout = 0
            # await asyncio.sleep(random.uniform(0.2, 2))

        # Wait for a message with the standard ACL template for negotiating to arrive.
        msg = await self.receive(timeout=timeout)  # Timeout set to 0s to continuously execute the behavior
        if msg:
            if msg.get_metadata(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_ATTRIB) == FIPAACLInfo.FIPA_ACL_PERFORMATIVE_PROPOSE:
                # A PROPOSE ACL message has been received by the agent
                _logger.aclinfo("         + PROPOSE Message received on SMIA Agent (HandleNegotiationBehaviour "
                                "in charge of the negotiation with thread [" + self.neg_thread + "])")
                _logger.aclinfo("                 |___ Message received with content: {}".format(msg.body))

                # The msg body will be parsed to a JSON object
                propose_msg_body_json = acl_smia_messages_utils.get_parsed_body_from_acl_msg(msg)

                # The negotiation information is obtained from the message
                # criteria = msg_json_body['serviceData']['serviceParams']['criteria']
                sender_agent_neg_value = propose_msg_body_json['negValue']

                if self.negotiation_result is None:
                    # Only if the negotiation is not resolved by this agent are the proposed messages analyzed
                    # The value of this SMIA and the received value are compared
                    if float(sender_agent_neg_value) > self.neg_value:
                        # As the received value is higher than this SMIA value, it must exit the negotiation.
                        self.negotiation_result = {'winner': False, 'timestamp': GeneralUtils.get_current_timestamp()}
                    if float(sender_agent_neg_value) == self.neg_value:
                        # In this case the negotiation is tied, so it must be managed
                        if not await self.handle_neg_values_tie(acl_smia_messages_utils.get_sender_from_acl_msg(msg),
                                                                float(sender_agent_neg_value)):
                            self.negotiation_result = {'winner': False,
                                                       'timestamp': GeneralUtils.get_current_timestamp()}

                # The target is added as processed in the local object (as it is a Python 'set' object there is no
                # problem of duplicate agents)
                self.targets_processed.add(acl_smia_messages_utils.get_sender_from_acl_msg(msg))

                if (len(self.targets_processed) == len(self.all_targets_list)) and self.negotiation_result is None:
                    # In this case all the values have already been received and the result is None, so the value of
                    # this SMIA is the best
                    _logger.info("The SMIA has won the negotiation with thread [" + msg.thread + "]")

                    # As the winner, it will reply to the sender with the result of the negotiation
                    inform_acl_msg = await inter_smia_interactions_utils.create_acl_smia_message(
                        self.received_body_json['negRequester'], self.neg_thread,
                        FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM,
                        self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB),
                        protocol=FIPAACLInfo.FIPA_ACL_CONTRACT_NET_PROTOCOL,
                        msg_body={'winner': True})
                    await self.send(inform_acl_msg)
                    _logger.aclinfo("ACL response sent for the result of the negotiation request with thread ["
                                    + msg.thread + "]")

                    # The negotiation can be terminated, in this case being the winner
                    self.negotiation_result = {'winner': True, 'timestamp': GeneralUtils.get_current_timestamp()}

            elif msg.get_metadata(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_ATTRIB) == FIPAACLInfo.FIPA_ACL_PERFORMATIVE_REQUEST:
                # In this case some action is requested within the negotiation
                request_msg_body_json = acl_smia_messages_utils.get_parsed_body_from_acl_msg(msg)
                if request_msg_body_json['requestedData'] == 'negValue':
                    sender_agent_jid = acl_smia_messages_utils.get_sender_from_acl_msg(msg)
                    _logger.aclinfo("         + REQUEST Message received on SMIA Agent (HandleNegotiationBehaviour) "
                                    "from " + sender_agent_jid + " with thread ["
                                    + self.neg_thread + "]) requesting the neg value")
                    # The neg value is sent with a PROPOSE message
                    self.targets_remaining.add(sender_agent_jid)
                    _logger.info("PROPOSE message sent with the requested neg value within thread [{}] to {}".format(
                        self.neg_thread, sender_agent_jid))

        else:
            # No ACL message has been received in this iteration
            if len(self.targets_remaining) == 0:
                # In this case, all PROPOSE messages have been sent
                if len(self.targets_processed) < len(self.all_targets_list):
                    if self.current_retries < self.max_retries:
                        # In this case, not all agents have been processed and there are still retries remaining
                        _logger.info("The negotiation with thread [{}] has not been resolved yet. Retry number {} "
                                     "requesting the negotiation value from the remaining targets."
                                     .format(self.neg_thread, self.current_retries + 1))
                        await self.request_remaining_neg_acl_msgs()
                        # 0.2 seconds are waited for each agent to receive messages
                        # await asyncio.sleep(0.2*(len(self.all_targets_list)-len(self.targets_processed)))
                        # await asyncio.sleep(random.uniform(0.2, 2))
                        # await asyncio.sleep(self.recovery_wait)
                        self.current_retries += 1
                else:
                    # In this case, all agents have been processed, so this agent has resolved the negotiation
                    if self.safe_iterations > 0:
                        # In safe iterations, it waits 0.1 second in case there is any agent that has not completed
                        # its processing
                        await asyncio.sleep(0.5)
                        self.safe_iterations -= 1
                    else:
                        if self.negotiation_result is not None:
                            # In this case the negotiation is resolved, so the behaviour can be killed
                            _logger.info("The negotiation with thread [{}] is resolved, so the behavior is killed"
                                         .format(self.neg_thread))
                            await self.exit_negotiation(is_winner=self.negotiation_result['winner'],
                                                        resolved_timestamp=self.negotiation_result['timestamp'])
                            return  # killing a behaviour does not cancel its current run loop
                        else:
                            # In this case  the negotiation is not resolved, so a failure message is sent to the requester
                            _logger.info(
                                "The negotiation with thread [{}] has not been resolved in {} safe iterations, so the "
                                "behavior is sending a 'FAILURE' message to the requester if it has not been resolved."
                                .format(self.neg_thread, self.safe_iterations))

                            missing_msgs = sum(
                                1 for jid_target in self.all_targets_list
                                if jid_target not in self.targets_processed
                            )

                            failure_acl_msg = await inter_smia_interactions_utils.create_acl_smia_message(
                                self.received_body_json['negRequester'], self.neg_thread,
                                FIPAACLInfo.FIPA_ACL_PERFORMATIVE_FAILURE,
                                self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB),
                                msg_body={
                                    'reason': "Negotiation not resolved ({} propose messages are not received).".format(
                                        missing_msgs), 'exceptionType': 'CapabilityRequestExecutionError',
                                    'affectedElement': self.received_body_json['capabilityIRI']})
                            await self.send(failure_acl_msg)
                            _logger.aclinfo("ACL failure message sent for the negotiation request with thread ["
                                            + self.neg_thread + "]")
                            await self.exit_negotiation(is_winner=False)
                            return  # killing a behaviour does not cancel its current run loop


    async def get_neg_value_with_criteria(self):
        """
        This method gets the negotiation value based on a given criteria.

        Returns:
            int: value of the negotiation
        """
        _logger.info("Getting the negotiation value for [{}]...".format(self.neg_thread))

        if self.neg_value != -1.0:

            # Since negotiation is a capability of the agent, it is necessary to analyze which skill has been defined.
            # The associated skill interface will be the one from which the value of negotiation can be obtained.
            # Thus, skill is the negotiation criterion for which the ontological instance will be obtained.
            neg_skill_instance = await self.myagent.css_ontology.get_ontology_instance_by_iri(
                self.received_body_json['negCriterion'])

            if neg_skill_instance is None:
                _logger.warning("This SMIA instance does not have the Skill Interface {} defined to obtain the "
                                "negotiation value. It will remain with the value 0.0.".format(
                    self.received_body_json['negCriterion']))
                return

            # The related skill interface will be obtained
            skill_interface = list(neg_skill_instance.get_associated_skill_interface_instances())[0]
            # The AAS element of the skill interface will be used to analyze the skill implementation
            aas_skill_interface_elem = await self.myagent.aas_model.get_object_by_reference(
                skill_interface.get_aas_sme_ref())

            parent_submodel = aas_skill_interface_elem.get_parent_submodel()
            if parent_submodel.check_semantic_id_exist(AssetInterfacesInfo.SEMANTICID_INTERFACES_SUBMODEL):
                # In this case, the value need to be obtained through an asset service
                # With the AAS SubmodelElement of the asset interface the related Python class, able to connect to the
                # asset, can be obtained.
                aas_asset_interface_elem = aas_skill_interface_elem.get_associated_asset_interface()
                asset_connection_class = await self.myagent.get_asset_connection_class_by_ref(aas_asset_interface_elem)
                _logger.assetinfo("The Asset connection of the Skill Interface has been obtained.")
                # Now the negotiation value can be obtained through an asset service
                _logger.assetinfo("Obtaining the negotiation value for [{}] through an asset service...".format(
                    self.neg_thread))
                current_neg_value = await asset_connection_class.execute_asset_service(
                    interaction_metadata=aas_skill_interface_elem)
                _logger.assetinfo("Negotiation value for [{}] through an asset service obtained: {}.".format(
                    self.neg_thread, current_neg_value))
                if not isinstance(current_neg_value, float):
                    try:
                        current_neg_value = float(current_neg_value)
                    except ValueError as e:
                        # TODO PENSAR OTRAS EXCEPCIONES EN NEGOCIACIONES (durante el asset connection...)
                        _logger.error(e)
                        raise CapabilityRequestExecutionError(self.neg_thread, 'Negotiation',
                                                              "The requested negotiation {} cannot be executed because the "
                                                              "negotiation value returned by the asset does not have a valid"
                                                              " format.".format(self.neg_thread), self)
            else:
                # In this case, the value need to be obtained through an agent service
                try:
                    current_neg_value = await self.myagent.agent_services.execute_agent_service_by_id(
                        aas_skill_interface_elem.id_short)
                except (KeyError, ValueError) as e:
                    _logger.error(e)
                    raise CapabilityRequestExecutionError(self.neg_thread, 'Negotiation',
                                                          "The requested negotiation {} cannot be executed because the "
                                                          "negotiation value cannot be obtained through the agent service "
                                                          "{}.".format(self.neg_thread,
                                                                       aas_skill_interface_elem.id_short), self)

            # Let's normalize the value between 0.0 and 1.0
            if 1.0 < current_neg_value <= 100.0:
                # Normalize within 0.0 and 100.0 range
                current_neg_value = (current_neg_value - 0.0) / 100.0
            self.neg_value = current_neg_value

    async def send_propose_acl_msgs(self, targets=None):
        """
        This method sends the FIPA-ACL messages with the PROPOSE performative in order to offer the negotiation value.
        If no targets are set, only a message is sent to distribute the communication load.

        Args:
            targets (list, optional): the targets to whom the proposal message should be sent.
        """

        if targets is None:
            if len(self.targets_remaining) == 0:
                return
            targets = [self.targets_remaining.pop()]

        propose_acl_message = await inter_smia_interactions_utils.create_acl_smia_message(
            'N/A', self.neg_thread,
            FIPAACLInfo.FIPA_ACL_PERFORMATIVE_PROPOSE,
            self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB),
            protocol=FIPAACLInfo.FIPA_ACL_CONTRACT_NET_PROTOCOL,
            msg_body={**self.received_body_json, **{'negValue': self.neg_value}})

        # This PROPOSE FIPA-ACL message is sent to all participants of the negotiation (except for this SMIA)
        for jid_target in targets:
            if jid_target != str(self.myagent.jid):
                propose_acl_message.to = jid_target
                await self.send(propose_acl_message)
                _logger.aclinfo("ACL PROPOSE negotiation message sent to " + jid_target +
                                "with neg value " + str(self.neg_value) +
                                " on negotiation with thread [" + self.neg_thread + "]")
                # await asyncio.sleep(0.01)   # It waits 0.01 second for each agent involved
                # await asyncio.sleep(random.uniform(0.005, 0.025))   # It waits random seconds for each agent. 5–25 ms, with enough jitter to break synchronization.
                # await asyncio.sleep(random.uniform(0.015, 0.035))   # It waits random seconds for each agent. 5–25 ms, with enough jitter to break synchronization.
                # await asyncio.sleep(random.uniform(0.025, 0.045))   # It waits random seconds for each agent. 5–25 ms, with enough jitter to break synchronization.
                await asyncio.sleep(self.time_sleep)   # Calculated sleep time for each sent message

    async def request_remaining_neg_acl_msgs(self):
        """
        This method sends the FIPA-ACL messages with a request for the PROPOSE message in order to obtain their
        negotiation value.
        """
        request_acl_message = await inter_smia_interactions_utils.create_acl_smia_message(
            'N/A', self.neg_thread,
            FIPAACLInfo.FIPA_ACL_PERFORMATIVE_REQUEST,
            self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB),
            protocol=FIPAACLInfo.FIPA_ACL_CONTRACT_NET_PROTOCOL,
            msg_body={'requestedData': 'negValue'})

        # This REQUEST FIPA-ACL message is sent to all participants of the negotiation (except for this SMIA)
        for jid_target in self.all_targets_list:
            if (jid_target not in self.targets_processed) and (jid_target != str(self.myagent.jid)):
                request_acl_message.to = jid_target
                await self.send(request_acl_message)
                _logger.aclinfo("ACL REQUEST negotiation message sent to " + jid_target +
                                "requesting the neg value on thread [" + self.neg_thread + "]")
                # await asyncio.sleep(0.01)    # It waits 0.01 second for each agent involved
                # await asyncio.sleep(random.uniform(0.005, 0.025))   # It waits random seconds for each agent. 5–25 ms, with enough jitter to break synchronization.
                # await asyncio.sleep(random.uniform(0.015, 0.035))   # It waits random seconds for each agent. 5–25 ms, with enough jitter to break synchronization.
                # await asyncio.sleep(random.uniform(0.025, 0.045))   # It waits random seconds for each agent. 5–25 ms, with enough jitter to break synchronization.
                await asyncio.sleep(self.time_sleep)  # Calculated sleep time for each sent message

    async def handle_neg_values_tie(self, received_agent_id, received_neg_value):
        """
        This method handles the situations where negotiation values tie. A seeded randomization process will be
        performed which will slightly modify the tied trading values and obtain a random winner. This method will be
        executed in all SMIA instances where the tie occurs, but since the ACL message thread is used as seed, they
        will all return the same result.

        Args:
            received_agent_id (str): identifier of the received SMIA agent proposal with the tie.
            received_neg_value (float): received tie negotiation value
        """
        if self.neg_value == -1.0 and received_neg_value == -1.0:
            # If both agents have not passed the Capability Check, neither of them can be the winner
            return False
        # First, the dictionary will be created with the agents that have the same negotiation value
        scores_dict = {str(self.myagent.jid): self.neg_value, received_agent_id: received_neg_value}
        # The pseudo-random number generator (PRNG) with the seed will give the same random values
        random.seed(self.neg_thread)
        # As the negotiation values are the same, the following will be disturbed
        perturbations = {opt: random.uniform(-0.001, 0.001)
               for opt in sorted(scores_dict.keys())}
        scores_dict_disturbed = {opt: scores_dict[opt] * (1 + perturbations[opt])
                                 for opt in perturbations}

        if max(scores_dict_disturbed, key=lambda k: scores_dict_disturbed[k]) != str(self.myagent.jid):
            # In this case the SMIA instance has loosened the negotiation, so a False is returned
            return False
        return True


    async def exit_negotiation(self, is_winner, resolved_timestamp=None):
        """
        This method is executed when the trade has ended, either as a winner or a loser. In any case, all the
        information of the negotiation is added to the global variable with all the information of all the negotiations
         of the agent. The thread is used to differentiate the information of each negotiation, since this is the
         identifier of each one of them.

        Args:
            is_winner (bool): it determines whether the SMIA has been the winner of the negotiation.
            resolved_timestamp (int): timestamp when the negotiation is resolved.
        """
        if is_winner:
            _logger.info("The SMIA has finished the negotiation with thread [" + self.neg_thread +
                         "] as the winner")
        else:
            _logger.info("The SMIA has finished the negotiation with thread [" + self.neg_thread +
                         "] not as the winner")

        # The negotiation information is stored in the global object of the SMIA
        neg_data_json = negotiation_utils.create_neg_json_to_store(neg_requester_jid=self.received_body_json['negRequester'],
                                                                   participants=self.received_body_json['negTargets'],
                                                                   neg_criteria=self.received_body_json['negCriterion'],
                                                                   neg_value=self.neg_value, is_winner=is_winner)
        await self.myagent.save_negotiation_data(thread=self.neg_thread, neg_data=neg_data_json)

        # The information will be stored in the log
        if resolved_timestamp is None:
            resolved_timestamp = GeneralUtils.get_current_timestamp()
        smia_archive_utils.save_completed_svc_log_info(self.requested_timestamp, resolved_timestamp,
                                                       await inter_smia_interactions_utils.acl_message_to_json(self.received_acl_msg), {'winner': is_winner},
                                                       self.received_acl_msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB))

        # Since this behavior is specific to the messages in this thread, it releases it so that the generic
        # ACLHandlingBehavior can process them in the future.
        await self.myagent.remove_reserved_thread(self.neg_thread)

        # In order to correctly complete the negotiation process, this behavior is removed from the agent.
        self.kill(exit_code=10)



import json
import logging
from json import JSONDecodeError

from smia import GeneralUtils
from spade.behaviour import CyclicBehaviour

from smia.behaviours.specific_handle_behaviours.handle_negotiation_behaviour import HandleNegotiationBehaviour
from smia.logic import inter_smia_interactions_utils, acl_smia_messages_utils
from smia.logic.exceptions import RequestDataError, ServiceRequestExecutionError
from smia.utilities.fipa_acl_info import FIPAACLInfo, ACLSMIAJSONSchemas

_logger = logging.getLogger(__name__)


class NegotiatingBehaviour(CyclicBehaviour):
    """
    This class implements the behaviour that handles the negotiation requests made by other standardized SMIAs
    through ACL messages in the I4.0 System.
    """

    def __init__(self, agent_object):
        """
        The constructor method is rewritten to add the object of the agent
        Args:
            agent_object (spade.Agent): the SPADE agent object of the SMIA agent.
        """

        # The constructor of the inherited class is executed.
        super().__init__()

        # The SPADE agent object is stored as a variable of the behaviour class
        self.myagent = agent_object

    async def on_start(self):
        """
        This method implements the initialization process of this behaviour.
        """
        _logger.info("NegotiationBehaviour starting...")

    async def run(self):
        """
        This method implements the logic of the behaviour.
        """

        # Wait for a message with the standard ACL template for negotiating to arrive.
        msg = await self.receive(
            timeout=10)  # Timeout set to 10 seconds so as not to continuously execute the behavior.
        if msg:
            # An ACL message has been received by the agent
            _logger.aclinfo("         + Message received on SMIA (NegotiatingBehaviour) from {}".format(msg.sender))
            _logger.aclinfo("                 |___ Message received with content: {}".format(msg.body))

            # This behaviour only manages interactions related to negotiations (FIPA-CNP protocol)

            # First, the message body will be checked against the associated ontology JSON Schema (only if it is not an
            # inform message, since this type of message can be strings)
            if msg.get_metadata(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_ATTRIB) != FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM:
                try:
                    # The msg body will be parsed to a JSON object
                    msg_json_body = json.loads(msg.body)

                    await inter_smia_interactions_utils.check_received_request_data_structure(
                        msg_json_body, ACLSMIAJSONSchemas.JSON_SCHEMA_ACL_SMIA_ONTOLOGIES_MAP.get(
                            msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB)))
                except (RequestDataError, JSONDecodeError) as cap_request_error:
                    # The added data are not valid, so a Refuse message to the requester must be sent
                    if isinstance(cap_request_error, JSONDecodeError):
                        cap_request_error.message = f"JSON error: {str(cap_request_error)}"
                    svc_execution_error = ServiceRequestExecutionError(msg.thread, cap_request_error.message,
                                                                       msg.get_metadata(
                                                                           FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB), self)
                    await svc_execution_error.handle_service_execution_error()
                    _logger.warning(
                        "The sender [{}] has sent an message with thread [{}] that has invalid data, therefore "
                        "it has been informed with a Refuse ACL message".format(
                            acl_smia_messages_utils.get_sender_from_acl_msg(msg), msg.thread))
                    return  # The run method is terminated to restart checking for new messages

            # When the message content has been validated, the specific behavior will be added to the agent to handle
            # the required actions within the FIPA-CNP protocol
            handle_neg_template_propose = GeneralUtils.create_acl_template(
                performative=FIPAACLInfo.FIPA_ACL_PERFORMATIVE_PROPOSE,
                protocol=FIPAACLInfo.FIPA_ACL_CONTRACT_NET_PROTOCOL, thread=msg.thread)
            handle_neg_template_request = GeneralUtils.create_acl_template(
                performative=FIPAACLInfo.FIPA_ACL_PERFORMATIVE_REQUEST,
                protocol=FIPAACLInfo.FIPA_ACL_CONTRACT_NET_PROTOCOL, thread=msg.thread)
            handle_neg_template = handle_neg_template_propose | handle_neg_template_request
            # handle_neg_template = SMIAInteractionInfo.NEG_STANDARD_ACL_TEMPLATE_PROPOSE
            # handle_neg_template.thread = msg.thread
            specific_neg_handling_behaviour = HandleNegotiationBehaviour(self.agent, received_acl_msg=msg)
            self.myagent.add_behaviour(specific_neg_handling_behaviour, handle_neg_template)
            _logger.info("Specific behaviour added to the agent to handle the message with thread [{}].".format(
                msg.thread))


            # # TODO BORRAR, ES CODIGO DEL ENFOQUE ANTIGUO
            # # The msg body will be parsed to a JSON object
            # msg_json_body = json.loads(msg.body)
            #
            # # TODO quizas hay que comprobar que el thread no existe? Para hacer al igual que en la peticion de
            # #  servicios, utilizar el thread como identificador. Hay que pensarlo, ya que de este modo el thread
            # #  serviría para un unico servicio a un AAS
            #
            # # Depending on the performative of the message, the agent will have to perform some actions or others
            # match msg.get_metadata('performative'):
            #     case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_CFP:
            #         _logger.aclinfo("The agent has received a request to start a negotiation (CFP) with thread ["
            #                         + msg.thread + "]")
            #
            #         # Since the negotiation is an agent capability, it will be checked the validity of the received data
            #         cap_req_data = None
            #         try:
            #             cap_req_data = inter_smia_interactions_utils.create_svc_json_data_from_acl_msg(msg)
            #             await self.check_received_capability_request_data(cap_req_data)
            #         except RequestDataError as cap_request_error:
            #             # The data are not valid, so the requester must be informed.
            #             acl_msg = inter_smia_interactions_utils.create_inter_smia_response_msg(
            #                 receiver=acl_smia_messages_utils.get_sender_from_acl_msg(msg),
            #                 thread=msg.thread,
            #                 performative=FIPAACLInfo.FIPA_ACL_PERFORMATIVE_FAILURE,
            #                 ontology=FIPAACLInfo.FIPA_ACL_ONTOLOGY_SVC_NEGOTIATION,   # TODO pensar si definir un NegRequest y NegResponse
            #                 service_id=msg_json_body['serviceID'],
            #                 service_type=msg_json_body['serviceType'],
            #                 service_params=json.dumps({'reason': cap_request_error.message})
            #             )
            #             await self.send(acl_msg)
            #             _logger.warning("Capability request with thread [{}] has invalid data, therefore the requester"
            #                             " has been informed".format(msg.thread))
            #             return  # The run method is terminated to restart checking for new messages.
            #
            #         # First, some useful information is obtained from the msg
            #         targets_list = cap_req_data['serviceData']['serviceParams']['targets'].split(',')
            #         neg_requester_jid = acl_smia_messages_utils.get_sender_from_acl_msg(msg)
            #         neg_criteria = cap_req_data['serviceData']['serviceParams'][CapabilitySkillACLInfo.REQUIRED_SKILL_NAME]
            #         if len(targets_list) == 1:
            #             # There is only one target available (therefore, it is the only one, so it is the winner)
            #             _logger.info("The SMIA has won the negotiation with thread [" + msg.thread + "]")
            #
            #             # As the winner, it will reply to the sender with the result of the negotiation
            #             acl_response_msg = negotiation_utils.create_neg_response_msg(
            #                 receiver=neg_requester_jid,
            #                 thread=msg.thread,
            #                 service_id=msg_json_body['serviceID'],
            #                 service_type=msg_json_body['serviceType'],
            #                 winner=str(self.myagent.jid))
            #             await self.send(acl_response_msg)
            #             _logger.aclinfo("ACL response sent for the result of the negotiation request with thread ["
            #                             + msg.thread + "]")
            #
            #             # The information will be stored in the log
            #             execution_info = {'capName': 'Negotiation',
            #                               'capType': CapabilitySkillOntologyUtils.AGENT_CAPABILITY_TYPE,
            #                               'participants': str(self.myagent.jid), 'criteria': neg_criteria,
            #                               'winner': 'True'}
            #             smia_archive_utils.save_completed_svc_log_info(
            #                 GeneralUtils.get_current_timestamp(), GeneralUtils.get_current_timestamp(),
            #                 inter_smia_interactions_utils.create_svc_json_data_from_acl_msg(msg), execution_info,
            #                 ServiceTypes.CSS_RELATED_SERVICE)
            #
            #             # Finally, the data is stored in the SMIA
            #             # TODO pensar si esto es necesario (ya se almacena todo en el JSON del Archive)
            #             neg_data_json = negotiation_utils.create_neg_json_to_store(
            #                 neg_requester_jid=neg_requester_jid,
            #                 participants=msg_json_body['serviceData']['serviceParams']['targets'],
            #                 neg_criteria=neg_criteria,
            #                 is_winner=True)
            #             await self.myagent.save_negotiation_data(thread=msg.thread, neg_data=neg_data_json)
            #
            #         else:
            #             # If there are more targets, a management behavior is added to the SMIA, which will be
            #             # in charge of this particular negotiation. To this end, some information has to be passed to
            #             # the behaviour
            #             behaviour_info = {
            #                 'thread': msg.thread,
            #                 'neg_requester_jid': neg_requester_jid,
            #                 'targets': msg_json_body['serviceData']['serviceParams']['targets'],
            #                 'neg_criteria': neg_criteria
            #             }
            #             # The FIPA-ACL template added to this behavior ensures that you will only receive PROPOSE
            #             # messages but also only with the thread of that specific thread
            #             handle_neg_template = SMIAInteractionInfo.NEG_STANDARD_ACL_TEMPLATE_PROPOSE
            #             handle_neg_template.thread = msg.thread
            #             neg_req_data = inter_smia_interactions_utils.create_svc_json_data_from_acl_msg(msg)
            #             handle_neg_behav = HandleNegotiationBehaviour(self.agent, behaviour_info, neg_req_data)
            #             self.myagent.add_behaviour(handle_neg_behav, handle_neg_template)
            #
            #     case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM:
            #         _logger.aclinfo("The agent has received an Inter AAS interaction message related to a negotiation:"
            #                         " Inform")
            #         if msg_json_body['serviceID'] == 'negotiationResult':
            #             _logger.aclinfo("The agent has received the result of the negotiation with thread ["+
            #                             msg.thread+"].")
            #             # TODO BORRAR (enfoque antiguo): pensar como gestionar las respuestas de
            #             # As the result of a negotiation is a response to a previous request, a new
            #             # HandleSvcResponseBehaviour to handle this service response will be added to the agent
            #             svc_resp_data = inter_smia_interactions_utils.create_svc_json_data_from_acl_msg(msg)
            #             svc_resp_handling_behav = HandleSvcResponseBehaviour(self.agent,
            #                                                                  'Inter AAS interaction',
            #                                                                  svc_resp_data)
            #             self.myagent.add_behaviour(svc_resp_handling_behav)
            #         else:
            #             _logger.warning('serviceID not available')
            #
            #     case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_FAILURE:
            #         _logger.aclinfo("The agent has received a response in a negotiation: Failure")
            #         # TODO pensar como se deberian gestionar este tipo de mensajes en una negociacion
            #     case _:
            #         _logger.error("ACL performative type not available.")
        else:
            _logger.info("         - No message received within 10 seconds on SMIA (NegotiatingBehaviour)")


    # async def check_received_capability_request_data(self, cap_req_data):
    #     """
    #     This method checks whether the data received contains the necessary information to be able to execute
    #     the agent capability. If an error occurs, it throws a CapabilityDataError exception.
    #
    #     Args:
    #         cap_req_data: (dict): all the information about the agent capability request
    #     """
    #     # First, the structure and attributes of the received data are checked and validated
    #     await inter_smia_interactions_utils.check_received_request_data_structure_old(
    #         cap_req_data, ACLSMIAJSONSchemas.JSON_SCHEMA_CAPABILITY_REQUEST)
    #     received_cap_data = cap_req_data['serviceData']['serviceParams']
    #     cap_name = received_cap_data[CapabilitySkillACLInfo.REQUIRED_CAPABILITY_NAME]
    #     cap_ontology_instance = await self.myagent.css_ontology.get_ontology_instance_by_name(cap_name)
    #     if cap_ontology_instance is None:
    #         raise RequestDataError("The capability {} does not an instance defined in the ontology of this "
    #                                "DT".format(cap_name))
    #     if CapabilitySkillACLInfo.REQUIRED_SKILL_NAME not in received_cap_data:
    #         raise RequestDataError("The skill need to be defined in negotiation requests.")
    #     skill_name = received_cap_data[CapabilitySkillACLInfo.REQUIRED_SKILL_NAME]
    #     result, skill_instance = cap_ontology_instance.check_and_get_related_instance_by_instance_name(skill_name)
    #     if result is False:
    #         raise RequestDataError("The capability {} and skill {} are not linked in the ontology of this "
    #                                "DT, or the skill does not have an instance"
    #                                ".".format(cap_name, skill_name))
    #     if CapabilitySkillACLInfo.REQUIRED_SKILL_INTERFACE_NAME in received_cap_data:
    #         # Solo si se ha definido la skill se define la skill interface, sino no tiene significado
    #         # TODO pensar la frase anterior. Realmente tiene significado o no? Si no se define la skill, podriamos
    #         #  definir una interfaz que queremos utilizar si o si? En ese caso, habria que buscar una skill con esa
    #         #  interfaz para ejecutarla
    #         skill_interface_name = received_cap_data[CapabilitySkillACLInfo.REQUIRED_SKILL_INTERFACE_NAME]
    #         result, instance = skill_instance.check_and_get_related_instance_by_instance_name(skill_interface_name)
    #         if result is False:
    #             raise RequestDataError("The skill {} and skill interface {} are not linked in the ontology of "
    #                                    "this DT, or the skill interface does not have an instance"
    #                                    ".".format(skill_name, skill_interface_name))
    #     # Negotiation requests require some mandatory values
    #     if 'targets' not in received_cap_data:
    #         raise RequestDataError("The negotiation capability request cannot be fulfilled due to lack of objectives "
    #                                "in the negotiation data.")


class AASRelatedServicesInfo:

    # The SMIA ISM identifier should always be the same, but if it changes, it should be changed here as well.
    SMIA_ISM_ID = 'smia-ism'

    # AAS Services information
    # ------------------------
    # AAS Services types
    AAS_SERVICE_TYPE_SUBMODEL_REGISTRY = 'SubmodelRegistryService'
    AAS_SERVICE_TYPE_METAINFORMATION = 'MetaInformationManagementService'
    AAS_SERVICE_TYPE_DISCOVERY = 'DiscoveryService'

    # AAS Discovery Services identifiers
    AAS_DISCOVERY_SERVICE_GET_SM_BY_REF = 'GetSubmodelElementByReference'
    AAS_DISCOVERY_SERVICE_GET_SM_VALUE_BY_REF = 'GetSubmodelElementValueByReference'
    AAS_DISCOVERY_SERVICE_GET_AAS_INFO = 'GetAASInformationElement'
    AAS_DISCOVERY_SERVICE_GET_SM_BY_ID = 'GetSubmodelBySubmodelID'
    AAS_DISCOVERY_SERVICE_SET_SM_VALUE_BY_REF = 'SetSubmodelElementValueByReference'

    # AAS Infrastructure Services information
    # ---------------------------------------
    # AAS Infrastructure Services types
    AAS_INFRASTRUCTURE_SERVICE_TYPE_REGISTRY = 'RegistryService'
    AAS_INFRASTRUCTURE_SERVICE_TYPE_DISCOVERY = 'DiscoveryService'

    # AAS Infrastructure Registry Services identifiers
    AAS_INFRASTRUCTURE_REGISTRY_SERVICE_REGISTER_SMIA = 'RegisterSMIAInstance'
    AAS_INFRASTRUCTURE_REGISTRY_CSS_ELEMENTS= 'RegisterCSSElements'

    # AAS Infrastructure Discovery Services identifiers
    AAS_INFRASTRUCTURE_DISCOVERY_SERVICE_GET_SMIA_BY_ASSET = 'GetSMIAInstanceIDByAssetID'
    AAS_INFRASTRUCTURE_DISCOVERY_SERVICE_GET_ASSET_BY_SMIA = 'GetAssetIDBySMIAInstanceID'
    AAS_INFRASTRUCTURE_DISCOVERY_SERVICE_GET_ALL_ASSET_BY_CAPABILITY = 'GetAllAssetIDByCapability'
    AAS_INFRASTRUCTURE_DISCOVERY_SERVICE_GET_AAS_BY_ID = 'GetAssetAdministrationShellById'




import logging

from smia.css_ontology.css_ontology_utils import CapabilitySkillOntologyInfo

import smia

from smia import GeneralUtils

from smia.utilities.fipa_acl_info import FIPAACLInfo, ACLSMIAOntologyInfo, ACLSMIAJSONSchemas
from spade.behaviour import State

from smia.behaviours.init_aas_model_behaviour import InitAASModelBehaviour
from smia.logic import inter_smia_interactions_utils, acl_smia_messages_utils
from smia.utilities import smia_archive_utils, properties_file_utils
from smia.utilities.aas_related_services_info import AASRelatedServicesInfo
from smia.utilities.general_utils import SMIAGeneralInfo

_logger = logging.getLogger(__name__)


class StateBooting(State):
    """
    This class contains the Boot state of the common SMIA.
    """

    async def run(self):
        """
        This method implements the boot state of the common SMIA. Here all the required initialization tasks
        are performed.
        """

        await self.booting_state_logic()
        self.set_next_state(SMIAGeneralInfo.RUNNING_STATE_NAME)

    async def booting_state_logic(self):
        """
        This method contains the logic of the boot state of the common SMIA. This method can be used by any
        inherited class.
        """
        _logger.info("## STATE 1: BOOTING ##  (Initial state)")

        # First, it is ensured that the attributes of the SMIA are initialized
        self.agent.initialize_smia_attributes()

        # The ontology has to be initialized in order to be available during the AAS model analysis
        await self.agent.css_ontology.initialize_ontology()

        # The submodels also have to be initialized, so its behaviour is also added
        init_aas_model_behav = InitAASModelBehaviour(self.agent)
        self.agent.add_behaviour(init_aas_model_behav)

        # Wait until the behaviours have finished because the AAS Archive has to be initialized to pass to running state
        await init_aas_model_behav.join()

        # If the initialization behaviour has completed, SMIA is in the InitializationReady status
        smia_archive_utils.update_status('InitializationReady')

        # When all initialization tasks have been completed, if the SMIA has been configured to register on the
        # SMIA-I KB, it will try to register through an infrastructure service provided by the SMIA ISM
        if str(properties_file_utils.get_dt_general_property('smia-i-kb-registration')).lower() in ('yes', 'true', 't', '1'):
            if await acl_smia_messages_utils.get_agent_id_from_jid(self.agent.jid) != AASRelatedServicesInfo.SMIA_ISM_ID:
                # The extracted CSS elements in the self-configuration process will be registered in the SMIA KB
                await self.send_register_css_elements_acl_msg()

                # The SMIA instance itself will be also registered in the SMIA KB
                await self.send_register_instance_acl_msg()

        # Finished the Boot State the agent can move to the next state
        _logger.info(f"{self.agent.jid} agent has finished it Boot state.")


    async def send_register_css_elements_acl_msg(self):
        """
        This method sends the ACL message to register all the CSS elements of the instance, extracted from the
        self-configuration process. All the CSS elements will be registered in the SMIA KB  through an infrastructure
        service provided by the SMIA ISM. It will wait 5 seconds for the 'inform' performative message with the
        registration confirmation. If it does not receive any message, it will just show a warning message and continue
        its execution.
        """
        css_elements_json = {'capabilities': [], 'skills': []}
        # First, the JSON of each capability will be obtained
        for capability in (filter(None,
                (await self.agent.css_ontology.get_ontology_instances_by_class_iri(
                    CapabilitySkillOntologyInfo.CSS_ONTOLOGY_CAPABILITY_IRI) or []) +
                (await self.agent.css_ontology.get_ontology_instances_by_class_iri(
                    CapabilitySkillOntologyInfo.CSS_ONTOLOGY_AGENT_CAPABILITY_IRI) or []) +
                (await self.agent.css_ontology.get_ontology_instances_by_class_iri(
                    CapabilitySkillOntologyInfo.CSS_ONTOLOGY_ASSET_CAPABILITY_IRI) or []))):
            capability_json = capability.from_ontology_instance_to_json()   # The JSON is created from OWL instance
            # The asset information is also added
            capability_json['assets'] = [{
                'id': await self.agent.aas_model.get_asset_information_attribute_value('asset_id'),
                'kind': await self.agent.aas_model.get_asset_information_attribute_value('asset_kind'),
                'type': await self.agent.aas_model.get_asset_information_attribute_value('asset_type')}]
            css_elements_json['capabilities'].append(capability_json)

        # Then, the JSON of each skill will be obtained
        for skill in (filter(None, (await self.agent.css_ontology.get_ontology_instances_by_class_iri(
                CapabilitySkillOntologyInfo.CSS_ONTOLOGY_SKILL_IRI) or []))):
            # The JSON is created from OWL instance
            css_elements_json['skills'].append(skill.from_ontology_instance_to_json())


        smia_ism_jid = (f"{AASRelatedServicesInfo.SMIA_ISM_ID}@"
                        f"{await acl_smia_messages_utils.get_xmpp_server_from_jid(self.agent.jid)}")
        register_acl_msg = await inter_smia_interactions_utils.create_acl_smia_message(
            # 'gcis1@xmpp.jp', await acl_smia_messages_utils.create_random_thread(self.agent),
            smia_ism_jid, await acl_smia_messages_utils.create_random_thread(self.agent),
            FIPAACLInfo.FIPA_ACL_PERFORMATIVE_REQUEST, ACLSMIAOntologyInfo.ACL_ONTOLOGY_AAS_INFRASTRUCTURE_SERVICE,
            protocol=FIPAACLInfo.FIPA_ACL_REQUEST_PROTOCOL, msg_body=await acl_smia_messages_utils.
            generate_json_from_schema(ACLSMIAJSONSchemas.JSON_SCHEMA_AAS_INFRASTRUCTURE_SERVICE, serviceID=
            AASRelatedServicesInfo.AAS_INFRASTRUCTURE_REGISTRY_CSS_ELEMENTS, serviceType=
            AASRelatedServicesInfo.AAS_INFRASTRUCTURE_SERVICE_TYPE_REGISTRY, serviceParams=css_elements_json))
        _logger.info("Sending the infrastructure service request to {} to register all the CSS elements of the SMIA "
                     "instance in the SMIA KB.".format(smia_ism_jid))
        await self.send(register_acl_msg)
        _logger.info("Waiting for the confirmation of the registry in the SMIA KB...")
        msg = await self.receive(timeout=3000)  # TODO DELETE: CHANGED FOR REGISTRATION PERFORMANCE TEST
        # msg = await self.receive(timeout=10)  # Timeout set to 10 seconds
        if msg:
            valid_msg_template = GeneralUtils.create_acl_template(
                FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM, ACLSMIAOntologyInfo.ACL_ONTOLOGY_AAS_INFRASTRUCTURE_SERVICE)
            if valid_msg_template.match(msg) and msg.thread == register_acl_msg.thread:
                _logger.aclinfo("CSS elements successfully registered in the SMIA KB")
            else:
                ("A message arrived but it is not about the CSS elements. Sender [{}], Performative [{}], "
                 "Ontology [{}]".format(acl_smia_messages_utils.get_sender_from_acl_msg(msg),
                                        msg.get_metadata(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_ATTRIB),
                                        msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB)))
        else:
            _logger.warning("The registry of CSS elements has not been completed. Check the SMIA ISM or the SMIA KB.")

    async def send_register_instance_acl_msg(self):
        """
        This method sends the ACL message to register its instance in the SMIA KB  through an infrastructure service
        provided by the SMIA ISM. It will wait 5 seconds for the 'inform' performative message with the registration
        confirmation. If it does not receive any message, it will just show a warning message and continue its execution.
        """
        smia_instance_json = {'id': str(self.agent.jid), 'asset':
            {'id': await self.agent.aas_model.get_asset_information_attribute_value('asset_id'),
             'kind': await self.agent.aas_model.get_asset_information_attribute_value('asset_kind'),
             'type': await self.agent.aas_model.get_asset_information_attribute_value('asset_type')},
                              'aasID':await self.agent.aas_model.get_aas_attribute_value('id'),
                              'status': 'Running', 'startedTimeStamp': GeneralUtils.get_current_timestamp(),
                              'smiaVersion': smia.__version__}
        smia_ism_jid = (f"{AASRelatedServicesInfo.SMIA_ISM_ID}@"
                        f"{await acl_smia_messages_utils.get_xmpp_server_from_jid(self.agent.jid)}")
        register_acl_msg = await inter_smia_interactions_utils.create_acl_smia_message(
            # 'gcis1@xmpp.jp', await acl_smia_messages_utils.create_random_thread(self.agent),
            smia_ism_jid, await acl_smia_messages_utils.create_random_thread(self.agent),
            FIPAACLInfo.FIPA_ACL_PERFORMATIVE_REQUEST, ACLSMIAOntologyInfo.ACL_ONTOLOGY_AAS_INFRASTRUCTURE_SERVICE,
            protocol=FIPAACLInfo.FIPA_ACL_REQUEST_PROTOCOL, msg_body=await acl_smia_messages_utils.
            generate_json_from_schema(ACLSMIAJSONSchemas.JSON_SCHEMA_AAS_INFRASTRUCTURE_SERVICE, serviceID=
            AASRelatedServicesInfo.AAS_INFRASTRUCTURE_REGISTRY_SERVICE_REGISTER_SMIA, serviceType=
            AASRelatedServicesInfo.AAS_INFRASTRUCTURE_SERVICE_TYPE_REGISTRY, serviceParams=smia_instance_json))
        _logger.info("Sending the infrastructure service request to {} to register the SMIA instance in the SMIA "
                     "KB.".format(smia_ism_jid))
        await self.send(register_acl_msg)
        _logger.info("Waiting for the confirmation of the registry in the SMIA KB...")
        msg = await self.receive(timeout=3000)  # TODO DELETE: CHANGED FOR REGISTRATION PERFORMANCE TEST
        # msg = await self.receive(timeout=10)  # Timeout set to 10 seconds
        _logger.assetinfo("HA PASADO DEL RECEIVE")  # TODO BORRAR
        if msg:
            _logger.assetinfo("MENSAJE RECEIVE: {}".format(msg))  # TODO BORRAR
            valid_msg_template = GeneralUtils.create_acl_template(
                FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM, ACLSMIAOntologyInfo.ACL_ONTOLOGY_AAS_INFRASTRUCTURE_SERVICE)
            if valid_msg_template.match(msg) and msg.thread == register_acl_msg.thread:
                _logger.aclinfo("SMIA instance successfully registered in the SMIA KB")

                # TODO DELETE: REGISTRATION PERFORMANCE TEST
                from smia.utilities import smia_archive_utils, smia_general_info
                import os
                registered_time = GeneralUtils.get_current_timer_nanosecs()
                metrics_folder = os.environ.get('METRICS_FOLDER')
                if metrics_folder is None:
                    metrics_folder = smia_general_info.SMIAGeneralInfo.CONFIGURATION_AAS_FOLDER_PATH + '/metrics'
                if self.agent.setup_time is None:
                    _logger.error("The registration time could not be obtained because the agent does not have the "
                                  "specified start time.")
                else:
                    await smia_archive_utils.save_csv_calculated_metrics(metrics_folder, self.agent.jid,
                                                                         registered_time-self.agent.setup_time,
                                                                         'Time elapsed for SMIA registration')
                    _logger.assetinfo("ALMACENADO TIEMPO {} en {}".format(metrics_folder,
                                                                          registered_time-self.agent.setup_time))  # TODO BORRAR

            else:
                ("A message arrived but it is not about the registration. Sender [{}], Performative [{}], "
                 "Ontology [{}]".format(acl_smia_messages_utils.get_sender_from_acl_msg(msg),
                                        msg.get_metadata(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_ATTRIB),
                                        msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB)))
        else:
            _logger.warning("The registry of the SMIA instance {} has not been completed. Check the SMIA ISM or the "
                            "SMIA KB.".format(str(self.agent.jid)))



import ast
import json
import random
import string

from smia.logic.exceptions import RequestDataError
from smia.logic.inter_smia_interactions_utils import check_received_request_data_structure, _logger

# Methods related to agent JIDs
# -----------------------------
async def get_xmpp_server_from_jid(agent_object_jid):
    """
    This method get the XMPP server from the JID of the agent.

    Args:
        agent_object_jid: the JID object of the SMIA SPADE agent.

    Returns:
        str: XMPP server.
    """
    try:
        return str(agent_object_jid.domain)
    except AttributeError:
        pass
    try:
        return str(agent_object_jid).split('@')[1]
    except AttributeError:
        pass
    return None

async def get_agent_id_from_jid(agent_object_jid):
    """
    This method get the identifier of the agent from the JID of the agent.

    Args:
        agent_object_jid: the JID object of the SMIA SPADE agent.

    Returns:
        str: XMPP server.
    """
    try:
        return str(agent_object_jid.localpart)
    except AttributeError:
        pass
    try:
        return str(agent_object_jid).split('@')[0]
    except AttributeError:
        pass
    return None


def get_sender_from_acl_msg(acl_msg):
    """
    This method returns the identifier of an agent from an ACL message, considering the suffixes that can be added
    by the XMPP server.

    Args:
        acl_msg (spade.message.Message): ACL message object.

    Returns:
        str: identifier of the sender of the message.
    """
    if '/' in str(acl_msg.sender):  # XMPP server can add a random string to differentiate the agent JID
        return str(acl_msg.sender).split('/')[0]
    else:
        return str(acl_msg.sender)

# Methods related to thread of the ACL messages
# ---------------------------------------------
async def create_random_thread(agent_object):
    """
    This method creates a value for the thread of a SPADE-ACL message, using the agent identifier and random string.

    Args:
        agent_object (spade.Agent): the SPADE agent object of the SMIA agent.

    Returns:
        str: thread value
    """
    return (f"{await get_agent_id_from_jid(agent_object.jid)}-"
            f"{''.join(random.choices(string.ascii_letters + string.digits, k=5)).lower()}")

# Methods related to body of the ACL messages
# -------------------------------------------
async def generate_json_from_schema(schema: dict, **kwargs) -> dict:
    """
    This method generates a valid JSON from a predefined JSON Schema.

    Args:
        schema (dict): JSON schema object.
        **kwargs: attributes along with their values to build the JSON.

    Returns:
        dict: valid JSON object regarding the given schema.
    """
    required_fields = schema.get("required", [])
    properties = schema.get("properties", {})
    json_object = {}

    # Mandatory fields
    for field in required_fields:
        if field not in kwargs:
            raise ValueError(f"Missing required parameter: {field}")
        if kwargs[field] is None:
            raise ValueError(f"Missing required value for parameter: {field}")
        json_object[field] = kwargs[field]

    # Optional fields
    for field in properties:
        if field not in json_object and field in kwargs and kwargs[field] is not None:
            json_object[field] = kwargs[field]

    # Validate final message
    try:
        await check_received_request_data_structure(json_object, schema)
    except RequestDataError as e:
        _logger.warning('A JSON object cannot be created using the schema. Check the failed code.')
        return None

    return json_object

def get_parsed_body_from_acl_msg(acl_msg):
    """
    This method gets the body of an ACL message and returns parsed.

    Args:
        acl_msg (spade.message.Message): the ACL message object.

    Returns:
        parsed body of the ACL message.
    """
    # Let's try with JSON
    try:

        return json.loads(acl_msg.body)
    except (json.JSONDecodeError, TypeError):
        pass
    # Now let's try Python literal evaluation, to safely evaluate Python literals (list, tuple, int, etc.))
    try:
        return ast.literal_eval(acl_msg.body)
    except (ValueError, SyntaxError):
        pass

    return acl_msg.body




class FIPAACLInfo:
    """
    This class contains the values related to FIPA-ACL standard.
    """

    # FIPA-ACL attribute values
    FIPA_ACL_PERFORMATIVE_ATTRIB = 'performative'
    FIPA_ACL_ONTOLOGY_ATTRIB = 'ontology'
    FIPA_ACL_PROTOCOL_ATTRIB = 'protocol'
    FIPA_ACL_ENCODING_ATTRIB = 'encoding'
    FIPA_ACL_LANGUAGE_ATTRIB = 'language'

    # Performative values
    FIPA_ACL_PERFORMATIVE_CFP = 'cfp'
    FIPA_ACL_PERFORMATIVE_INFORM = 'inform'
    FIPA_ACL_PERFORMATIVE_REQUEST = 'request'
    FIPA_ACL_PERFORMATIVE_PROPOSE = 'propose'
    FIPA_ACL_PERFORMATIVE_FAILURE = 'failure'
    FIPA_ACL_PERFORMATIVE_QUERY_IF = 'query-if'
    FIPA_ACL_PERFORMATIVE_QUERY_REF = 'query-ref'
    FIPA_ACL_PERFORMATIVE_REFUSE = 'refuse'
    FIPA_ACL_PERFORMATIVE_NOT_UNDERSTOOD = 'not-understood'
    FIPA_ACL_PERFORMATIVE_ACCEPT_PROPOSAL = 'accept-proposal'
    FIPA_ACL_PERFORMATIVE_REJECT_PROPOSAL = 'reject-proposal'

    # Protocol values
    FIPA_ACL_REQUEST_PROTOCOL = 'fipa-request'
    FIPA_ACL_CONTRACT_NET_PROTOCOL = 'fipa-contract-net'
    FIPA_ACL_QUERY_PROTOCOL = 'fipa-query'

    # TODO borrar: Antiguos nombres (se han actualizado a los formatos definidos en FIPA (en minuscula))
    # FIPA_ACL_PERFORMATIVE_CFP = 'CallForProposal'
    # FIPA_ACL_PERFORMATIVE_INFORM = 'Inform'
    # FIPA_ACL_PERFORMATIVE_REQUEST = 'Request'
    # FIPA_ACL_PERFORMATIVE_PROPOSE = 'Propose'
    # FIPA_ACL_PERFORMATIVE_FAILURE = 'Failure'
    # FIPA_ACL_PERFORMATIVE_QUERY_IF = 'Query-If'
    # TODO add more if they are needed
    # TODO se han añadido estos pero todavia no se utilizan:
    FIPA_ACL_PERFORMATIVE_ACCEPT_PROPOSAL = 'AcceptProposal'
    FIPA_ACL_PERFORMATIVE_REJECT_PROPOSAL = 'RejectProposal'
    FIPA_ACL_PERFORMATIVE_AGREE = 'Agree'
    FIPA_ACL_PERFORMATIVE_CONFIRM = 'Confirm'
    FIPA_ACL_PERFORMATIVE_NOT_UNDERSTOOD = 'NotUnderstood'
    FIPA_ACL_PERFORMATIVE_REFUSE = 'Refuse'

    # Ontology values
    FIPA_ACL_ONTOLOGY_SVC_REQUEST = 'SvcRequest'
    FIPA_ACL_ONTOLOGY_SVC_RESPONSE = 'SvcResponse'
    FIPA_ACL_ONTOLOGY_CAPABILITY_REQUEST = 'CapabilityRequest'
    FIPA_ACL_ONTOLOGY_CAPABILITY_CHECKING = 'CapabilityChecking'
    FIPA_ACL_ONTOLOGY_CAPABILITY_RESPONSE = 'CapabilityResponse'
    FIPA_ACL_ONTOLOGY_SVC_NEGOTIATION = 'Negotiation'

    # Default values in SMIA approach for some attributes
    FIPA_ACL_DEFAULT_ENCODING = 'application/json'
    FIPA_ACL_DEFAULT_LANGUAGE = 'smia-language'

class ACLSMIAOntologyInfo:
    """
    This class contains the values related to ACL-SMIA ontology library information: the ontology classification
    established for the SMIA approach.
    """
    ACL_ONTOLOGY_ASSET_RELATED_SERVICE = 'asset-related-service'
    ACL_ONTOLOGY_AGENT_RELATED_SERVICE = 'agent-related-service'
    ACL_ONTOLOGY_AAS_SERVICE = 'aas-service'
    ACL_ONTOLOGY_AAS_INFRASTRUCTURE_SERVICE = 'aas-infrastructure-service'
    ACL_ONTOLOGY_CSS_SERVICE = 'css-service'

class ServiceTypes:
    """
    This class contains all service types defined in the Functional View of RAMI 4.0.
    """
    ASSET_RELATED_SERVICE = 'AssetRelatedService'
    AAS_INFRASTRUCTURE_SERVICE = 'AASInfrastructureService'
    AAS_SERVICE = 'AASservice'
    SUBMODEL_SERVICE = 'SubmodelService'
    CSS_RELATED_SERVICE = 'CSSRelatedService'   # TODO duda con este ya que contiene el concepto de service dentro, y no es lo mismo

class ACLSMIAJSONSchemas:
    """This class contains all the JSON schemas related to ACL messages sent between SMIA agents."""

    JSON_SCHEMA_AAS_MODEL_REFERENCE_old = {
        "type": "object",
        "properties": {
            "keys": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "type": {"type": "string"},
                        "value": {"type": "string"}
                    },
                    "required": ["type", "value"]
                }
            }
        },
        "required": ["keys"]
    }

    JSON_SCHEMA_SUBMODEL_SERVICE_REQUEST = {
        "type": "object",
        "properties": {
            "ModelReference": JSON_SCHEMA_AAS_MODEL_REFERENCE_old,
            "ExternalReference": {
                "type": "string"
            }
        },
        "oneOf": [
            {"required": ["ModelReference"]},
            {"required": ["ExternalReference"]}
        ]
    }

    JSON_SCHEMA_ASSET_SERVICE_REQUEST = {
        "type": "object",
        "properties": {
            "serviceParameterValues": {
                "type": "object",
                "additionalProperties": {"type": "string"}
            },
            "ModelReference": JSON_SCHEMA_AAS_MODEL_REFERENCE_old,
        },
        "required": ["ModelReference"]
    }

    JSON_SCHEMA_CAPABILITY_REQUEST = {
        "type": "object",
        "properties": {
            "capabilityName": {"type": "string"},
            "skillName": {"type": "string"},
            "skillParameterValues": {
              "type": "object",
              "additionalProperties": {
                "type": "string"
              }
            },
            "skillInterfaceName": {"type": "string"}
        },
        "required": ["capabilityName"]
    }

    # NEW SCHEMAS FOR SMIA: added in v0.2.4
    # -------------------------------------
    # -------------------------------------

    # Common schemas
    # --------------
    AAS_MODEL_REFERENCE_STRING_PATTERN = "^(\\[[^\\[\\],]+,[^\\[\\],]+\\])+$"
    JSON_SCHEMA_AAS_MODEL_REFERENCE = {
        "oneOf": [
            {
                "type": "object",
                "properties": {
                    "keys": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "type": {"type": "string"},
                                "value": {"type": "string"}
                            },
                            "required": ["type", "value"]
                        }
                    }
                },
                "required": ["keys"]
            },
            {
                "type": "string",
                "pattern": AAS_MODEL_REFERENCE_STRING_PATTERN,
                "description": "ModelReference format for strings: [type,value][type,value]..."
            }
        ]
    }

    JSON_SCHEMA_SMIA_INSTANCE_INFORMATION = {
        "type": "object",
        "properties": {
            "id": { "type": "string" },
            "asset": {
                "type": "object",
                "properties": {
                    "id": {"type": "string"},
                    "kind": {"type": "string"},
                    "type": {"type": "string"},
                }
            },
            "aasID": {"type": "string"},
            "status": {"type": "string"},
            "startedTimeStamp": {"type": "integer"},
            "smiaVersion": {"type": "string"},
        },
        "required": ["id", "asset", "aasID"],
    }
    # TODO pensar mas (ademas de añadir ModelReference mas adelante)

    # Asset-related services / agent-related services schemas
    # -------------------------------------------------------
    JSON_SCHEMA_ASSET_AGENT_RELATED_SERVICE = {
        "title": "AssetAgentRelatedServiceSchema",
        "type": "object",

        "properties": {
            "serviceRef": JSON_SCHEMA_AAS_MODEL_REFERENCE,
            "serviceParams": {
                "type": "object",
                "additionalProperties": {"type": "string"}
            }
        },
        "required": ["serviceRef"]
    }
    # TODO falta probarlos

    # AAS Services schemas
    # --------------------
    JSON_SCHEMA_AAS_SERVICE = {
        # "$schema": "http://json-schema.org/draft-07/schema#",
        "title": "AASServiceSchema",
        "type": "object",

        "properties": {
            "serviceID": {"type": "string"},
            "serviceType": {
                "type": "string",
                "enum": ["SubmodelRegistryService", "MetaInformationManagementService", "DiscoveryService"]  # TODO in the future think about adding more
            },
            "serviceParams": {}
        },
        "required": ["serviceID", "serviceType"],
        "allOf": [
            {
                "if": {
                    "properties": {"serviceID": {"const": "GetSubmodelElementByReference"}}
                },
                "then": {
                    "properties": {
                        "serviceParams": JSON_SCHEMA_AAS_MODEL_REFERENCE
                    }
                }
            },
            {
                "if": {
                    "properties": {
                        "serviceID": {
                            # TODO Think of more AAS services that simply require a string.
                            "enum": ["GetAASInformationByAssetID"]
                        }
                    }
                },
                "then": {
                    "properties": {
                        "serviceParams": {"type": "string"}
                    }
                }
            }
        ]
    }
    # TODO falta probarlos

    # AAS Infrastructure Services schemas
    # -----------------------------------
    JSON_SCHEMA_AAS_INFRASTRUCTURE_SERVICE = {
        # "$schema": "http://json-schema.org/draft-07/schema#",
        "title": "AASInfrastructureServiceSchema",
        "type": "object",

        "properties": {
            "serviceID": { "type": "string" },
            "serviceType": {
                "type": "string",
                "enum": ["RegistryService", "DiscoveryService"]   # TODO in the future think about adding more
            },
        "serviceParams": {}
        },
        "required": ["serviceID", "serviceType"],
        "allOf": [
            {
                "if": {
                    "properties": { "serviceID": { "const": "RegistrySMIAInstance" } }
                },
                "then": {
                    "properties": { "serviceParams": JSON_SCHEMA_SMIA_INSTANCE_INFORMATION }
                }
            },
            {
                "if": {
                    "properties": {
                        "serviceID": {
                            # TODO Think of more infrastructure services that simply require a string.
                            "enum": ["GetSMIAInstanceIDByAssetID", "GetAssetIDBySMIAInstanceID",
                                     "GetAllAssetIDByCapability"]
                        }
                    }
                },
                "then": {
                    "properties": {
                        "serviceParams": { "type": "string" }
                    }
                }
            }
        ]
    }
    # TODO falta probarlos

    # CSS Services schemas
    # --------------------
    JSON_SCHEMA_CSS_SERVICE = {
        "type": "object",
        "properties": {
            "capabilityIRI": {"type": "string"},
            "skillIRI": {"type": "string"},
            "constraints": {
                "type": "object",
                "additionalProperties": {"type": "string"}
            },
            "skillParams": {
                "type": "object",
                "additionalProperties": {"type": "string"}
            },
            "skillInterfaceIRI": {"type": "string"},

            # Optional parameters for SMIA instances negotiation within FIPA-CNP protocol
            "negCriterion": {"type": "string"},
            "negTargets": {
                "type": "array",
                "items": {"type": "string"}
            },
            "negRequester": {"type": "string"},
            "negValue": {
                "type": "number",   # It is a float number between 0 and 1
                "minimum": 0.0,
                "maximum": 1.0
            },
        },
        "required": ["capabilityIRI"]
    }
    # TODO falta probarlos


    # JSON schemas mapping to ACL-SMIA ontologies
    # -------------------------------------------
    JSON_SCHEMA_ACL_SMIA_ONTOLOGIES_MAP = {
        ACLSMIAOntologyInfo.ACL_ONTOLOGY_ASSET_RELATED_SERVICE: JSON_SCHEMA_ASSET_AGENT_RELATED_SERVICE,
        ACLSMIAOntologyInfo.ACL_ONTOLOGY_AGENT_RELATED_SERVICE: JSON_SCHEMA_ASSET_AGENT_RELATED_SERVICE,
        ACLSMIAOntologyInfo.ACL_ONTOLOGY_AAS_SERVICE: JSON_SCHEMA_AAS_SERVICE,
        ACLSMIAOntologyInfo.ACL_ONTOLOGY_AAS_INFRASTRUCTURE_SERVICE: JSON_SCHEMA_AAS_INFRASTRUCTURE_SERVICE,
        ACLSMIAOntologyInfo.ACL_ONTOLOGY_CSS_SERVICE: JSON_SCHEMA_CSS_SERVICE
    }


import json
import logging
from json import JSONDecodeError

from smia.logic.exceptions import RequestDataError, ServiceRequestExecutionError
from spade.behaviour import CyclicBehaviour

from smia.behaviours.specific_handle_behaviours.handle_capability_behaviour import HandleCapabilityBehaviour
from smia.behaviours.specific_handle_behaviours.handle_aas_related_svc_behaviour import HandleAASRelatedSvcBehaviour
from smia.logic import inter_smia_interactions_utils, acl_smia_messages_utils
from smia.utilities.fipa_acl_info import FIPAACLInfo, ACLSMIAJSONSchemas, ACLSMIAOntologyInfo

_logger = logging.getLogger(__name__)


class ACLHandlingBehaviour(CyclicBehaviour):
    """
    This class implements the behaviour that handles all the ACL messages that the SMIA will receive from the
    others standardized SMIAs in the I4.0 System.
    """

    def __init__(self, agent_object):
        """
        The constructor method is rewritten to add the object of the agent.

        Args:
            agent_object (spade.Agent): the SPADE agent object of the SMIA agent.
        """

        # The constructor of the inherited class is executed.
        super().__init__()

        # The SPADE agent object is stored as a variable of the behaviour class
        self.myagent = agent_object

    async def on_start(self):
        """
        This method implements the initialization process of this behaviour.
        """
        _logger.info("ACLHandlingBehaviour starting...")

    async def run(self):
        """
        This method implements the logic of the behaviour.
        """
        try:
            # Wait for a message with the standard ACL template to arrive.
            msg = await self.receive(
                timeout=10)  # Timeout set to 10 seconds so as not to continuously execute the behavior.
            if msg:
                # An ACL message has been received by the agent
                _logger.aclinfo("         + Message received on SMIA (ACLHandlingBehaviour) from {}".format(msg.sender))
                _logger.aclinfo("                 |___ Message received with content: {}".format(msg.body))

                # If the ontology value is not within the valid values in the SMIA approach, a 'not understood' ACL message
                # is returned.
                if (msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB) not in
                        ACLSMIAJSONSchemas.JSON_SCHEMA_ACL_SMIA_ONTOLOGIES_MAP.keys()):
                    await inter_smia_interactions_utils.send_response_msg_from_received(
                        self, msg, FIPAACLInfo.FIPA_ACL_PERFORMATIVE_NOT_UNDERSTOOD)
                    return  # The run method is terminated to restart checking for new messages

                # Only if the thread is not reserved it needs to be processed
                if msg.thread not in self.myagent.reserved_threads:
                    # First, the message body will be checked against the associated ontology JSON Schema (only if it is not an
                    # inform message, since this type of message can be strings)
                    if msg.get_metadata(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_ATTRIB) != FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM:
                        try:
                            # The msg body will be parsed to a JSON object
                            msg_json_body = json.loads(msg.body)

                            await inter_smia_interactions_utils.check_received_request_data_structure(
                                msg_json_body, ACLSMIAJSONSchemas.JSON_SCHEMA_ACL_SMIA_ONTOLOGIES_MAP.get(
                                    msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB)))
                        except (RequestDataError, JSONDecodeError) as cap_request_error:
                            # The added data are not valid, so a Refuse message to the requester must be sent
                            if isinstance(cap_request_error, JSONDecodeError):
                                cap_request_error.message = f"JSON error: {str(cap_request_error)}"
                            svc_execution_error = ServiceRequestExecutionError(msg.thread, cap_request_error.message,
                                                                               msg.get_metadata(
                                                                                   FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB), self)
                            await svc_execution_error.handle_service_execution_error()
                            _logger.warning("The sender [{}] has sent an message with thread [{}] that has invalid data, therefore "
                                            "it has been informed with a Refuse ACL message".format(
                                acl_smia_messages_utils.get_sender_from_acl_msg(msg), msg.thread))
                            return  # The run method is terminated to restart checking for new messages

                    # Depending on the message ontology, the associated specific behavior will be added to the agent to handle
                    # the required actions
                    match msg.get_metadata(FIPAACLInfo.FIPA_ACL_ONTOLOGY_ATTRIB):
                        case ACLSMIAOntologyInfo.ACL_ONTOLOGY_CSS_SERVICE:
                            specific_handling_behaviour = HandleCapabilityBehaviour(self.myagent, received_acl_msg=msg)
                        case _:
                            specific_handling_behaviour = HandleAASRelatedSvcBehaviour(self.myagent, received_acl_msg=msg)
                    self.myagent.add_behaviour(specific_handling_behaviour)
                    _logger.info("Specific behaviour added to the agent to handle the message with thread [{}].".format(
                        msg.thread))

                    # # TODO BORRAR, ES CODIGO DEL ENFOQUE ANTIGUO
                    # # Depending on the performative of the message, the agent will have to perform some actions or others
                    # match msg.get_metadata('performative'):
                    #     case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_REQUEST:
                    #         _logger.aclinfo("The performative of the message is Request, so the DT needs to perform some"
                    #                         " action.")
                    #
                    #         # TODO pensar si generar un behaviour para recibir peticiones de servicios y otro para peticiones
                    #         #  de capacidades (en ese caso solo hay que comprobar la ontologia, no la performativa, ya que esta
                    #         #  se analiza en los behaviours de gestion individuales)
                    #         if ((msg.get_metadata('ontology') == FIPAACLInfo.FIPA_ACL_ONTOLOGY_CAPABILITY_REQUEST) or
                    #                 (msg.get_metadata('ontology') == FIPAACLInfo.FIPA_ACL_ONTOLOGY_CAPABILITY_CHECKING)):
                    #             _logger.aclinfo("The agent has received a request related to Capability-Skill model")
                    #             # The behaviour to handle this specific capability will be added to the agent
                    #             svc_req_data = inter_smia_interactions_utils.create_svc_json_data_from_acl_msg(msg)
                    #             capability_handling_behav = HandleCapabilityBehaviour(self.agent, svc_req_data)
                    #             self.myagent.add_behaviour(capability_handling_behav)
                    #
                    #         elif msg.get_metadata('ontology') == FIPAACLInfo.FIPA_ACL_ONTOLOGY_SVC_REQUEST:
                    #             _logger.aclinfo("The agent has received a request to perform a service")
                    #             # The behaviour to handle this specific capability will be added to the agent
                    #             svc_req_data = inter_smia_interactions_utils.create_svc_json_data_from_acl_msg(msg)
                    #             svc_req_handling_behav = HandleAASRelatedSvcBehaviour(self.agent, svc_req_data)
                    #             self.myagent.add_behaviour(svc_req_handling_behav)
                    #         else:
                    #             # TODO Parte del enfoque antiguo
                    #             service_category = msg_json_body['serviceData']['serviceCategory']
                    #             if service_category == 'service-response':
                    #                 # TODO PENSAR COMO RECOGER LAS RESPUESTAS DE PETICIONES ANTERIORES. NO VENDRIAN CON
                    #                 #  PERFORMATIVE 'Request', sai que aqui no se deberian tratar (recoger con Performative
                    #                 #  'Inform', 'Confirm', 'Agree', 'Refuse'... dentro de los behaviours para gestiones
                    #                 #  individuales?)
                    #                 _logger.aclinfo(
                    #                     "The agent has received the response of a service with thread [" + msg.thread + "].")
                    #
                    #                 # As it is a response to a previous request, a new HandleSvcResponseBehaviour to handle this service
                    #                 # response will be added to the agent
                    #                 svc_resp_data = inter_smia_interactions_utils.create_svc_json_data_from_acl_msg(msg)
                    #                 svc_resp_handling_behav = HandleSvcResponseBehaviour(self.agent,
                    #                                                                      'Inter AAS interaction',
                    #                                                                      svc_resp_data)
                    #                 self.myagent.add_behaviour(svc_resp_handling_behav)
                    #
                    #     case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_QUERY_IF:
                    #         _logger.aclinfo("The performative of the message is Query-If, so the DT has been asked about some "
                    #                         "aspect of it.")
                    #
                    #         # service_category = msg_json_body['serviceID']
                    #         # if service_category == 'capabilityChecking':
                    #         if msg.get_metadata('ontology') == FIPAACLInfo.FIPA_ACL_ONTOLOGY_CAPABILITY_CHECKING:
                    #             _logger.info("The DT has been asked to check if it has a given capability.")
                    #             svc_req_data = inter_smia_interactions_utils.create_svc_json_data_from_acl_msg(msg)
                    #             capability_handling_behav = HandleCapabilityBehaviour(self.agent, svc_req_data)
                    #             self.myagent.add_behaviour(capability_handling_behav)
                    #     case _:
                    #         _logger.error("ACL performative type not available.")

            else:
                _logger.info("         - No message received within 10 seconds on SMIA (ACLHandlingBehaviour)")
        except Exception as e:
            _logger.error("ERROR in ACLHandlingBehaviour: {}".format(e))



from smia.css_ontology.css_ontology_utils import CapabilitySkillACLInfo


async def capability_checking(agent_object, received_css_json):
    """
    This method executes the Capability Checking operation. In this step, the capabilities offered by this SMIA are
    matched against the received request. In the future, reallocation logic can be applied to infer the capacity
    request received, or advanced techniques such as auto-discovery can be adopted to discover capacities automatically.

    Args:
        agent_object (smia.agents.SMIAAgent): SMIA agent object.
        received_css_json (dict): received CSS data within the SMIA-ACL body message.

    Returns:
        bool, str: boolean with the result of the checking and, if false, reason of the fail
    """
    # First, the ontology instance of the capability is obtained
    cap_name = received_css_json[CapabilitySkillACLInfo.ATTRIB_CAPABILITY_IRI]
    cap_ontology_instance = await agent_object.css_ontology.get_ontology_instance_by_iri(cap_name)

    # The associated constraints and received values for them are also obtained
    constraint_instance_list = cap_ontology_instance.get_associated_constraint_instances()
    if constraint_instance_list is not None:
        received_constraint_data = received_css_json[CapabilitySkillACLInfo.ATTRIB_CAPABILITY_CONSTRAINTS]

        # At this point, the data received has already been verified, so that the values of the constraints can be
        # checked directly.
        for constraint_instance in constraint_instance_list:
            aas_cap_constraint_elem = await agent_object.aas_model.get_object_by_reference(
                constraint_instance.get_aas_sme_ref())
            result = aas_cap_constraint_elem.check_constraint(received_constraint_data[constraint_instance.iri])
            if not result:
                return False, 'The constraint {} with data {} is not valid'.format(constraint_instance.name,
                                                                                   received_constraint_data[
                                                                                       constraint_instance.iri])
    # If all capability constraint are valid, the checking valid
    # TODO PENSAR MAS VALIDACIONES DURANTE EL CAPABILITY CHECKING
    return True, ''



async def feasibility_checking(agent_object, received_css_json):
    """
    This method executes the Feasibility Checking operation. The objective of this step is to ensure that the necessary
     conditions hold so that it is feasible for the selected resource to perform its task. These conditions can be
     pre-conditions, post-conditions or invariants (hold over the entire duration).

    Args:
        agent_object (smia.agents.SMIAAgent): SMIA agent object.
        received_css_json (dict): received CSS data within the SMIA-ACL body message.

    Returns:
        bool, str: boolean with the result of the checking and, if false, reason of the fail
    """
    # TODO
    pass